# Copyright 2015 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
import os.path
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader, StdinShader


def simple_vertex_shader():
    return """#version 310 es
    void main() {
        gl_Position = vec4(1., 2., 3., 4.);
    }"""


def simple_fragment_shader():
    return """#version 310 es
    void main() {
        gl_FragDepth = 10.;
    }"""


def simple_compute_shader():
    return """#version 310 es
    void main() {
        uvec3 temp = gl_WorkGroupID;
    }"""


@inside_glslc_testsuite('OptionDashCapS')
class TestSingleDashCapSSingleFile(expect.ValidAssemblyFile):
    """Tests that -S works with a single file."""

    shader = FileShader(simple_vertex_shader(), '.vert')
    glslc_args = ['-S', shader]


@inside_glslc_testsuite('OptionDashCapS')
class TestSingleFileSingleDashCapS(expect.ValidAssemblyFile):
    """Tests that the position of -S doesn't matter."""

    shader = FileShader(simple_vertex_shader(), '.vert')
    glslc_args = [shader, '-S']


@inside_glslc_testsuite('OptionDashCapS')
class TestSingleDashCapSMultipleFiles(expect.ValidAssemblyFile):
    """Tests that -S works with multiple files."""

    shader1 = FileShader(simple_vertex_shader(), '.vert')
    shader2 = FileShader(simple_vertex_shader(), '.vert')
    shader3 = FileShader(simple_fragment_shader(), '.frag')
    glslc_args = ['-S', shader1, shader2, shader3]


@inside_glslc_testsuite('OptionDashCapS')
class TestMultipleDashCapSSingleFile(expect.ValidAssemblyFile):
    """Tests that multiple -Ss works as one."""

    shader = FileShader(simple_vertex_shader(), '.vert')
    glslc_args = ['-S', '-S', shader, '-S']


@inside_glslc_testsuite('OptionDashCapS')
class TestMultipleDashCapSMultipleFiles(expect.ValidAssemblyFile):
    """Tests a mix of -Ss and files."""

    shader1 = FileShader(simple_fragment_shader(), '.frag')
    shader2 = FileShader(simple_vertex_shader(), '.vert')
    shader3 = FileShader(simple_compute_shader(), '.comp')
    glslc_args = ['-S', shader1, '-S', '-S', shader2, '-S', shader3, '-S']


@inside_glslc_testsuite('OptionDashCapS')
class TestDashCapSWithDashC(expect.ValidAssemblyFile):
    """Tests that -S overwrites -c."""

    shader1 = FileShader(simple_fragment_shader(), '.frag')
    shader2 = FileShader(simple_vertex_shader(), '.vert')
    glslc_args = ['-c', '-S', shader1, '-c', '-c', shader2]


@inside_glslc_testsuite('OptionDashCapS')
class TestDashCapSWithDashFShaderStage(expect.ValidAssemblyFile):
    """Tests that -S works with -fshader-stage=."""

    shader1 = FileShader(simple_fragment_shader(), '.glsl')
    shader2 = FileShader(simple_vertex_shader(), '.glsl')
    shader3 = FileShader(simple_compute_shader(), '.glsl')
    glslc_args = ['-S',
                  '-fshader-stage=fragment', shader1,
                  '-fshader-stage=vertex', shader2,
                  '-fshader-stage=compute', shader3]


@inside_glslc_testsuite('OptionDashCapS')
class TestDashCapSWithDashStd(expect.ValidAssemblyFileWithWarning):
    """Tests that -S works with -std=."""

    shader1 = FileShader(simple_fragment_shader(), '.frag')
    shader2 = FileShader(simple_vertex_shader(), '.vert')
    shader3 = FileShader(simple_compute_shader(), '.comp')
    glslc_args = ['-S', '-std=450', shader1, shader2, shader3]

    w = (': warning: (version, profile) forced to be (450, none), '
         'while in source code it is (310, es)\n')
    expected_warning = [
        shader1, w, shader2, w, shader3, w, '3 warnings generated.\n']


@inside_glslc_testsuite('OptionDashCapS')
class TestDashCapSWithDashOSingleFile(expect.SuccessfulReturn,
                                      expect.CorrectAssemblyFilePreamble):
    """Tests that -S works with -o on a single file."""

    shader = FileShader(simple_fragment_shader(), '.frag')
    glslc_args = ['-S', '-o', 'blabla', shader]

    def check_output_blabla(self, status):
        output_name = os.path.join(status.directory, 'blabla')
        return self.verify_assembly_file_preamble(output_name)


@inside_glslc_testsuite('OptionDashCapS')
class TestDashCapSWithDashOMultipleFiles(expect.ErrorMessage):
    """Tests that -S works with -o on a single file."""

    shader1 = FileShader(simple_fragment_shader(), '.frag')
    shader2 = FileShader(simple_vertex_shader(), '.vert')
    glslc_args = ['-S', '-o', 'blabla', shader1, shader2]

    expected_error = ['glslc: error: cannot specify -o when '
                      'generating multiple output files\n']


@inside_glslc_testsuite('OptionDashCapS')
class TestDashCapSWithStdIn(expect.ValidAssemblyFile):
    """Tests that -S works with stdin."""

    shader = StdinShader(simple_fragment_shader())
    glslc_args = ['-S', '-fshader-stage=fragment', shader]


@inside_glslc_testsuite('OptionDashCapS')
class TestDashCapSWithStdOut(
    expect.ReturnCodeIsZero, expect.StdoutMatch, expect.StderrMatch):
    """Tests that -S works with stdout."""

    shader = FileShader(simple_fragment_shader(), '.frag')
    glslc_args = ['-S', '-o', '-', shader]

    expected_stdout = True
    expected_stderr = ''



# Copyright 2015 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
import os.path
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader, StdinShader, TempFileName


@inside_glslc_testsuite('File')
class SimpleFileCompiled(expect.ValidObjectFile):
    """Tests whether or not a simple glsl file compiles."""

    shader = FileShader('#version 310 es\nvoid main() {}', '.frag')
    glslc_args = ['-c', shader]


@inside_glslc_testsuite('File')
class NotSpecifyingOutputName(expect.SuccessfulReturn,
                              expect.CorrectObjectFilePreamble):
    """Tests that when there is no -o and -E/-S/-c specified, output as a.spv."""

    shader = FileShader('#version 140\nvoid main() {}', '.frag')
    glslc_args = [shader]

    def check_output_a_spv(self, status):
        output_name = os.path.join(status.directory, 'a.spv')
        return self.verify_object_file_preamble(output_name)


@inside_glslc_testsuite('Parameters')
class HelpParameters(
    expect.ReturnCodeIsZero, expect.StdoutMatch, expect.StderrMatch):
    """Tests the --help flag outputs correctly and does not produce and error."""

    glslc_args = ['--help']

    expected_stdout = '''glslc - Compile shaders into SPIR-V

Usage: glslc [options] file...

An input file of - represents standard input.

Options:
  -c                Only run preprocess, compile, and assemble steps.
  -Dmacro[=defn]    Add an implicit macro definition.
  -E                Outputs only the results of the preprocessing step.
                    Output defaults to standard output.
  -fauto-bind-uniforms
                    Automatically assign bindings to uniform variables that
                    don't have an explicit 'binding' layout in the shader
                    source.
  -fauto-map-locations
                    Automatically assign locations to uniform variables that
                    don't have an explicit 'location' layout in the shader
                    source.
  -fauto-combined-image-sampler
                    Removes sampler variables and converts existing textures
                    to combined image-samplers.
  -fentry-point=<name>
                    Specify the entry point name for HLSL compilation, for
                    all subsequent source files.  Default is "main".
  -fhlsl-16bit-types
                    Enable 16-bit type support for HLSL.
  -fhlsl_functionality1, -fhlsl-functionality1
                    Enable extension SPV_GOOGLE_hlsl_functionality1 for HLSL
                    compilation.
  -fhlsl-iomap      Use HLSL IO mappings for bindings.
  -fhlsl-offsets    Use HLSL offset rules for packing members of blocks.
                    Affects only GLSL.  HLSL rules are always used for HLSL.
  -finvert-y        Invert position.Y output in vertex shader.
  -flimit=<settings>
                    Specify resource limits. Each limit is specified by a limit
                    name followed by an integer value.  Tokens should be
                    separated by whitespace.  If the same limit is specified
                    several times, only the last setting takes effect.
  -flimit-file <file>
                    Set limits as specified in the given file.
  -fnan-clamp       Generate code for max and min builtins so that, when given
                    a NaN operand, the other operand is returned. Similarly,
                    the clamp builtin will favour the non-NaN operands, as if
                    clamp were implemented as a composition of max and min.
  -fpreserve-bindings
                    Preserve all binding declarations, even if those bindings
                    are not used.
  -fresource-set-binding [stage] <reg0> <set0> <binding0>
                        [<reg1> <set1> <binding1>...]
                    Explicitly sets the descriptor set and binding for
                    HLSL resources, by register name.  Optionally restrict
                    it to a single stage.
  -fcbuffer-binding-base [stage] <value>
                    Same as -fubo-binding-base.
  -fimage-binding-base [stage] <value>
                    Sets the lowest automatically assigned binding number for
                    images.  Optionally only set it for a single shader stage.
                    For HLSL, the resource register number is added to this
                    base.
  -fsampler-binding-base [stage] <value>
                    Sets the lowest automatically assigned binding number for
                    samplers  Optionally only set it for a single shader stage.
                    For HLSL, the resource register number is added to this
                    base.
  -fssbo-binding-base [stage] <value>
                    Sets the lowest automatically assigned binding number for
                    shader storage buffer objects (SSBO).  Optionally only set
                    it for a single shader stage.  Only affects GLSL.
  -ftexture-binding-base [stage] <value>
                    Sets the lowest automatically assigned binding number for
                    textures.  Optionally only set it for a single shader stage.
                    For HLSL, the resource register number is added to this
                    base.
  -fuav-binding-base [stage] <value>
                    For automatically assigned bindings for unordered access
                    views (UAV), the register number is added to this base to
                    determine the binding number.  Optionally only set it for
                    a single shader stage.  Only affects HLSL.
  -fubo-binding-base [stage] <value>
                    Sets the lowest automatically assigned binding number for
                    uniform buffer objects (UBO).  Optionally only set it for
                    a single shader stage.
                    For HLSL, the resource register number is added to this
                    base.
  -fshader-stage=<stage>
                    Treat subsequent input files as having stage <stage>.
                    Valid stages are vertex, vert, fragment, frag, tesscontrol,
                    tesc, tesseval, tese, geometry, geom, compute, and comp.
  -g                Generate source-level debug information.
  -h                Display available options.
  --help            Display available options.
  -I <value>        Add directory to include search path.
  -mfmt=<format>    Output SPIR-V binary code using the selected format. This
                    option may be specified only when the compilation output is
                    in SPIR-V binary code form. Available options are:
                      bin   - SPIR-V binary words.  This is the default.
                      c     - Binary words as C initializer list of 32-bit ints
                      num   - List of comma-separated 32-bit hex integers
  -M                Generate make dependencies. Implies -E and -w.
  -MM               An alias for -M.
  -MD               Generate make dependencies and compile.
  -MF <file>        Write dependency output to the given file.
  -MT <target>      Specify the target of the rule emitted by dependency
                    generation.
  -O                Optimize the generated SPIR-V code for better performance.
  -Os               Optimize the generated SPIR-V code for smaller size.
  -O0               Disable optimization.
  -o <file>         Write output to <file>.
                    A file name of '-' represents standard output.
  -std=<value>      Version and profile for GLSL input files. Possible values
                    are concatenations of version and profile, e.g. 310es,
                    450core, etc.  Ignored for HLSL files.
  -S                Emit SPIR-V assembly instead of binary.
  --show-limits     Display available limit names and their default values.
  --target-env=<environment>
                    Set the target client environment, and the semantics
                    of warnings and errors.  An optional suffix can specify
                    the client version.  Values are:
                        vulkan1.0       # The default
                        vulkan1.1
                        vulkan1.2
                        vulkan1.3
                        vulkan          # Same as vulkan1.0
                        opengl4.5
                        opengl          # Same as opengl4.5
  --target-spv=<spirv-version>
                    Set the SPIR-V version to be used for the generated SPIR-V
                    module.  The default is the highest version of SPIR-V
                    required to be supported for the target environment.
                    For example, default for vulkan1.0 is spv1.0, and
                    the default for vulkan1.1 is spv1.3,
                    the default for vulkan1.2 is spv1.5.
                    the default for vulkan1.3 is spv1.6.
                    Values are:
                        spv1.0, spv1.1, spv1.2, spv1.3, spv1.4, spv1.5, spv1.6
  --version         Display compiler version information.
  -w                Suppresses all warning messages.
  -Werror           Treat all warnings as errors.
  -x <language>     Treat subsequent input files as having type <language>.
                    Valid languages are: glsl, hlsl.
                    For files ending in .hlsl the default is hlsl.
                    Otherwise the default is glsl.
'''

    expected_stderr = ''


@inside_glslc_testsuite('Parameters')
class HelpIsNotTooWide(expect.StdoutNoWiderThan80Columns):
    """Tests that --help output is not too wide."""

    glslc_args = ['--help']


@inside_glslc_testsuite('Parameters')
class UnknownSingleLetterArgument(expect.ErrorMessage):
    """Tests that an unknown argument triggers an error message."""

    glslc_args = ['-a']
    expected_error = ["glslc: error: unknown argument: '-a'\n"]


@inside_glslc_testsuite('Parameters')
class UnknownMultiLetterArgument(expect.ErrorMessage):
    """Tests that an unknown argument triggers an error message."""

    glslc_args = ['-zzz']
    expected_error = ["glslc: error: unknown argument: '-zzz'\n"]


@inside_glslc_testsuite('Parameters')
class UnsupportedOption(expect.ErrorMessage):
    """Tests that an unsupported option triggers an error message."""

    glslc_args = ['--unsupported-option']
    expected_error = [
        "glslc: error: unsupported option: '--unsupported-option'\n"]


@inside_glslc_testsuite('File')
class FileNotFound(expect.ErrorMessage):
    """Tests the error message if a file cannot be found."""

    blabla_file = TempFileName('blabla.frag')
    glslc_args = [blabla_file]
    expected_error = [
        "glslc: error: cannot open input file: '", blabla_file,
        "': No such file or directory\n"]


@inside_glslc_testsuite('Unsupported')
class LinkingNotSupported(expect.ErrorMessage):
    """Tests the error message generated by linking not supported yet."""

    shader1 = FileShader('#version 140\nvoid main() {}', '.vert')
    shader2 = FileShader('#version 140\nvoid main() {}', '.frag')
    glslc_args = [shader1, shader2]
    expected_error = [
        'glslc: error: linking multiple files is not supported yet. ',
        'Use -c to compile files individually.\n']


@inside_glslc_testsuite('Unsupported')
class MultipleStdinUnsupported(expect.ErrorMessage):
    """Tests the error message generated by having more than one - input."""

    glslc_args = ['-c', '-fshader-stage=vertex', '-', '-']
    expected_error = [
        'glslc: error: specifying standard input "-" as input more'
        ' than once is not allowed.\n']


@inside_glslc_testsuite('Parameters')
class StdinWithoutShaderStage(expect.StdoutMatch, expect.StderrMatch):
    """Tests that you must use -fshader-stage when specifying - as input."""
    shader = StdinShader(
        """#version 140
    int a() {
    }
    void main() {
      int x = a();
    }
    """)
    glslc_args = [shader]

    expected_stdout = ''
    expected_stderr = [
        "glslc: error: '-': -fshader-stage required when input is from "
        'standard input "-"\n']


@inside_glslc_testsuite('Parameters')
class LimitsHelp(expect.StdoutMatch, expect.StderrMatch):
    """Tests --show-limits shows correct output."""

    glslc_args = ['--show-limits']

    expected_stderr = ''
    expected_stdout = """MaxLights 8
MaxClipPlanes 6
MaxTextureUnits 2
MaxTextureCoords 8
MaxVertexAttribs 16
MaxVertexUniformComponents 4096
MaxVaryingFloats 60
MaxVertexTextureImageUnits 16
MaxCombinedTextureImageUnits 80
MaxTextureImageUnits 16
MaxFragmentUniformComponents 1024
MaxDrawBuffers 8
MaxVertexUniformVectors 256
MaxVaryingVectors 15
MaxFragmentUniformVectors 256
MaxVertexOutputVectors 16
MaxFragmentInputVectors 15
MinProgramTexelOffset -8
MaxProgramTexelOffset 7
MaxClipDistances 8
MaxComputeWorkGroupCountX 65535
MaxComputeWorkGroupCountY 65535
MaxComputeWorkGroupCountZ 65535
MaxComputeWorkGroupSizeX 1024
MaxComputeWorkGroupSizeY 1024
MaxComputeWorkGroupSizeZ 64
MaxComputeUniformComponents 512
MaxComputeTextureImageUnits 16
MaxComputeImageUniforms 8
MaxComputeAtomicCounters 8
MaxComputeAtomicCounterBuffers 1
MaxVaryingComponents 60
MaxVertexOutputComponents 64
MaxGeometryInputComponents 64
MaxGeometryOutputComponents 128
MaxFragmentInputComponents 128
MaxImageUnits 8
MaxCombinedImageUnitsAndFragmentOutputs 8
MaxCombinedShaderOutputResources 8
MaxImageSamples 0
MaxVertexImageUniforms 0
MaxTessControlImageUniforms 0
MaxTessEvaluationImageUniforms 0
MaxGeometryImageUniforms 0
MaxFragmentImageUniforms 8
MaxCombinedImageUniforms 8
MaxGeometryTextureImageUnits 16
MaxGeometryOutputVertices 256
MaxGeometryTotalOutputComponents 1024
MaxGeometryUniformComponents 512
MaxGeometryVaryingComponents 60
MaxTessControlInputComponents 128
MaxTessControlOutputComponents 128
MaxTessControlTextureImageUnits 16
MaxTessControlUniformComponents 1024
MaxTessControlTotalOutputComponents 4096
MaxTessEvaluationInputComponents 128
MaxTessEvaluationOutputComponents 128
MaxTessEvaluationTextureImageUnits 16
MaxTessEvaluationUniformComponents 1024
MaxTessPatchComponents 120
MaxPatchVertices 32
MaxTessGenLevel 64
MaxViewports 16
MaxVertexAtomicCounters 0
MaxTessControlAtomicCounters 0
MaxTessEvaluationAtomicCounters 0
MaxGeometryAtomicCounters 0
MaxFragmentAtomicCounters 8
MaxCombinedAtomicCounters 8
MaxAtomicCounterBindings 1
MaxVertexAtomicCounterBuffers 0
MaxTessControlAtomicCounterBuffers 0
MaxTessEvaluationAtomicCounterBuffers 0
MaxGeometryAtomicCounterBuffers 0
MaxFragmentAtomicCounterBuffers 0
MaxCombinedAtomicCounterBuffers 1
MaxAtomicCounterBufferSize 32
MaxTransformFeedbackBuffers 4
MaxTransformFeedbackInterleavedComponents 64
MaxCullDistances 8
MaxCombinedClipAndCullDistances 8
MaxSamples 4
MaxMeshOutputVerticesNV 256
MaxMeshOutputPrimitivesNV 512
MaxMeshWorkGroupSizeX_NV 32
MaxMeshWorkGroupSizeY_NV 1
MaxMeshWorkGroupSizeZ_NV 1
MaxTaskWorkGroupSizeX_NV 32
MaxTaskWorkGroupSizeY_NV 1
MaxTaskWorkGroupSizeZ_NV 1
MaxMeshViewCountNV 4
MaxMeshOutputVerticesEXT 256
MaxMeshOutputPrimitivesEXT 256
MaxMeshWorkGroupSizeX_EXT 128
MaxMeshWorkGroupSizeY_EXT 128
MaxMeshWorkGroupSizeZ_EXT 128
MaxTaskWorkGroupSizeX_EXT 128
MaxTaskWorkGroupSizeY_EXT 128
MaxTaskWorkGroupSizeZ_EXT 128
MaxMeshViewCountEXT 4
MaxDualSourceDrawBuffersEXT 1
"""



# Copyright 2015 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader


def simple_vertex_shader():
    return """#version 310 es
    void main() {
        gl_Position = vec4(1., 2., 3., 4.);
    }"""


def simple_hlsl_vertex_shader():
    # Use "main" so we don't have to specify -fentry-point
    return """float4 main() : SV_POSITION { return float4(1.0); } """


def simple_fragment_shader():
    return """#version 310 es
    void main() {
        gl_FragDepth = 10.;
    }"""


def simple_tessellation_control_shader():
    return """#version 440 core
    layout(vertices = 3) out;
    void main() { }"""


def simple_tessellation_evaluation_shader():
    return """#version 440 core
    layout(triangles) in;
    void main() { }"""


def simple_geometry_shader():
    return """#version 150 core
    layout (triangles) in;
    layout (line_strip, max_vertices = 4) out;
    void main() { }"""


def simple_compute_shader():
    return """#version 310 es
    void main() {
        uvec3 temp = gl_WorkGroupID;
    }"""


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageWithGlslExtension(expect.ValidObjectFile):
    """Tests -fshader-stage with .glsl extension."""

    shader = FileShader(simple_vertex_shader(), '.glsl')
    glslc_args = ['-c', '-fshader-stage=vertex', shader]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageWithHlslExtension(expect.ValidObjectFile):
    """Tests -fshader-stage with .hlsl extension."""

    shader = FileShader(simple_hlsl_vertex_shader(), '.hlsl')
    glslc_args = ['-c', '-fshader-stage=vertex', shader]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageWithKnownExtension(expect.ValidObjectFile):
    """Tests -fshader-stage with known extension."""

    shader = FileShader(simple_fragment_shader(), '.frag')
    glslc_args = ['-c', '-fshader-stage=fragment', shader]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageWithUnknownExtension(expect.ValidObjectFile):
    """Tests -fshader-stage with unknown extension."""

    shader = FileShader(simple_vertex_shader(), '.unknown')
    glslc_args = ['-c', '-fshader-stage=vertex', shader]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageWithNoExtension(expect.ValidObjectFile):
    """Tests -fshader-stage with no extension."""

    shader = FileShader(simple_vertex_shader(), '')
    glslc_args = ['-c', '-fshader-stage=vertex', shader]


@inside_glslc_testsuite('OptionShaderStage')
class TestAllShaderStages(expect.ValidObjectFile):
    """Tests all possible -fshader-stage values."""

    shader1 = FileShader(simple_vertex_shader(), '.glsl')
    shader2 = FileShader(simple_fragment_shader(), '.glsl')
    shader3 = FileShader(simple_tessellation_control_shader(), '.glsl')
    shader4 = FileShader(simple_tessellation_evaluation_shader(), '.glsl')
    shader5 = FileShader(simple_geometry_shader(), '.glsl')
    shader6 = FileShader(simple_compute_shader(), '.glsl')
    glslc_args = [
        '-c',
        '-fshader-stage=vertex', shader1,
        '-fshader-stage=fragment', shader2,
        '-fshader-stage=tesscontrol', shader3,
        '-fshader-stage=tesseval', shader4,
        '-fshader-stage=geometry', shader5,
        '-fshader-stage=compute', shader6]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageOverwriteFileExtension(expect.ValidObjectFile):
    """Tests -fshader-stage has precedence over file extension."""

    # a vertex shader camouflaged with .frag extension
    shader = FileShader(simple_vertex_shader(), '.frag')
    # Command line says it's vertex shader. Should compile successfully
    # as a vertex shader.
    glslc_args = ['-c', '-fshader-stage=vertex', shader]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageLatterOverwriteFormer(expect.ValidObjectFile):
    """Tests a latter -fshader-stage overwrite a former one."""

    shader = FileShader(simple_vertex_shader(), '.glsl')
    glslc_args = [
        '-c', '-fshader-stage=fragment', '-fshader-stage=vertex', shader]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageWithMultipleFiles(expect.ValidObjectFile):
    """Tests -fshader-stage covers all subsequent files."""

    shader1 = FileShader(simple_vertex_shader(), '.glsl')
    # a vertex shader with .frag extension
    shader2 = FileShader(simple_vertex_shader(), '.frag')
    shader3 = FileShader(simple_vertex_shader(), '.a_vert_shader')
    glslc_args = ['-c', '-fshader-stage=vertex', shader1, shader2, shader3]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageMultipleShaderStage(expect.ValidObjectFile):
    """Tests multiple -fshader-stage."""

    shader1 = FileShader(simple_vertex_shader(), '.glsl')
    shader2 = FileShader(simple_fragment_shader(), '.frag')
    shader3 = FileShader(simple_vertex_shader(), '.a_vert_shader')
    glslc_args = [
        '-c',
        '-fshader-stage=vertex', shader1,
        '-fshader-stage=fragment', shader2,
        '-fshader-stage=vertex', shader3]


@inside_glslc_testsuite('OptionShaderStage')
class TestFileExtensionBeforeShaderStage(expect.ValidObjectFile):
    """Tests that file extensions before -fshader-stage are not affected."""

    # before -fshader-stage
    shader1 = FileShader(simple_vertex_shader(), '.vert')
    # after -fshader-stage
    shader2 = FileShader(simple_fragment_shader(), '.frag')
    shader3 = FileShader(simple_fragment_shader(), '.vert')
    glslc_args = ['-c', shader1, '-fshader-stage=fragment', shader2, shader3]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageWrongShaderStageValue(expect.ErrorMessage):
    """Tests that wrong shader stage value results in an error."""

    shader = FileShader(simple_vertex_shader(), '.glsl')
    glslc_args = ['-c', '-fshader-stage=unknown', shader]
    expected_error = ["glslc: error: stage not recognized: 'unknown'\n"]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageGlslExtensionMissingShaderStage(expect.ErrorMessage):
    """Tests that missing -fshader-stage for .glsl extension results in
    an error."""

    shader = FileShader(simple_vertex_shader(), '.glsl')
    glslc_args = ['-c', shader]
    expected_error = [
        "glslc: error: '", shader,
        "': .glsl file encountered but no -fshader-stage specified ahead\n"]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageHlslExtensionMissingShaderStage(expect.ErrorMessage):
    """Tests that missing -fshader-stage for .hlsl extension results in
    an error."""

    shader = FileShader(simple_hlsl_vertex_shader(), '.hlsl')
    glslc_args = ['-c', '-x', 'hlsl', shader]
    expected_error = [
        "glslc: error: '", shader,
        "': .hlsl file encountered but no -fshader-stage specified ahead\n"]


@inside_glslc_testsuite('OptionShaderStage')
class TestShaderStageUnknownExtensionMissingShaderStage(expect.ErrorMessage):
    """Tests that missing -fshader-stage for unknown extension results in
    an error."""

    shader = FileShader(simple_vertex_shader(), '.a_vert_shader')
    glslc_args = ['-c', shader]
    expected_error = [
        "glslc: error: '", shader,
        "': file not recognized: File format not recognized\n"]



# Copyright 2018 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader

# An HLSL shader with a counter buffer with a counter increment.
# Glslang doesn't automatically assign a binding to the counter, and
# it doesn't understand [[vk::counter_binding(n)]], so compile this
# with --auto-bind-uniforms.
# See https://github.com/KhronosGroup/glslang/issues/1616
HLSL_VERTEX_SHADER_WITH_COUNTER_BUFFER = """
RWStructuredBuffer<int> Ainc;
float4 main() : SV_Target0 {
  return float4(Ainc.IncrementCounter(), 0, 1, 2);
}
"""


@inside_glslc_testsuite('OptionFHlslFunctionality1')
class TestHlslFunctionality1MentionsExtension(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fhlsl_functionality1 enabled SPV_GOOGLE_hlsl_functionality1."""

    shader = FileShader(HLSL_VERTEX_SHADER_WITH_COUNTER_BUFFER, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl_functionality1',
                  '-fauto-bind-uniforms', shader]
    expected_assembly_substr = 'OpExtension "SPV_GOOGLE_hlsl_functionality1"'


@inside_glslc_testsuite('OptionFHlslFunctionality1')
class TestHlslFunctionality1DecoratesCounter(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fhlsl_functionality1 decorates the output target"""

    shader = FileShader(HLSL_VERTEX_SHADER_WITH_COUNTER_BUFFER, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl_functionality1',
                  '-fauto-bind-uniforms', shader]
    expected_assembly_substr = 'OpDecorateString'


## Next tests use the option with the hypen instead of underscore.

@inside_glslc_testsuite('OptionFHlslFunctionality1')
class TestHlslHyphenFunctionality1MentionsExtension(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fhlsl-functionality1 enabled SPV_GOOGLE_hlsl_functionality1."""

    shader = FileShader(HLSL_VERTEX_SHADER_WITH_COUNTER_BUFFER, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl_functionality1',
                  '-fauto-bind-uniforms', shader]
    expected_assembly_substr = 'OpExtension "SPV_GOOGLE_hlsl_functionality1"'


@inside_glslc_testsuite('OptionFHlslFunctionality1')
class TestHlslHyphenFunctionality1DecoratesCounter(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fhlsl-functionality1 decorates the output target"""

    shader = FileShader(HLSL_VERTEX_SHADER_WITH_COUNTER_BUFFER, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl_functionality1',
                  '-fauto-bind-uniforms', shader]
    expected_assembly_substr = 'OpDecorateString'



# Copyright 2017 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader

# A GLSL shader with uniforms without explicit bindings.
GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS = """#version 450
  #extension GL_ARB_sparse_texture2 : enable
  uniform texture2D my_tex;
  uniform sampler my_sam;
  layout(rgba32f) uniform image2D my_img;
  layout(rgba32f) uniform imageBuffer my_imbuf;
  uniform block { float x; float y; } my_ubo;
  buffer B { float z; } my_ssbo;
  void main() {
    texture(sampler2D(my_tex,my_sam),vec2(1.0));
    vec4 t;
    sparseImageLoadARB(my_img,ivec2(0),t);
    imageLoad(my_imbuf,42);
    float x = my_ubo.x;
    my_ssbo.z = 1.1;
  }"""


# An HLSL shader with uniforms without explicit bindings.
# The counter buffer has an associated counter that needs a
# binding.  Compile this shader with --auto-bind-uniforms to
# give it a binding, since Glslang does not support attribute
# [[vk::counter_binding(N))]].
# See https://github.com/KhronosGroup/glslang/issues/1616
HLSL_SHADER_WITHOUT_BINDINGS = """
SamplerState s1 : register(s1);
SamplerComparisonState s2 : register(s2);

Texture1D t1 : register(t11);
Texture2D <float4> t2 : register(t12);
Texture3D <float2> t3 : register(t13);
StructuredBuffer<float4> t4 : register(t14);
ByteAddressBuffer t5 : register(t15);
Buffer<float4> t6 : register(t16);

RWTexture1D <float4>  u1 : register(u21);
RWTexture1D <float4>  u2 : register(u22);
RWTexture1D <float4>  u3 : register(u23);
RWBuffer <float4> u4 : register(u24);
RWByteAddressBuffer u5 : register(u25);
RWStructuredBuffer<float> u6 : register(u26);
AppendStructuredBuffer<float> u7 : register(u27);
ConsumeStructuredBuffer<float> u8 : register(u28);

float4 main() : SV_Target0 {
   s1;
   s2;

   t1;
   t2;
   t3;
   t4[2];
   t5.Load(8);
   t6;

   u1;
   u2;
   u3;
   u4[1];
   u5.Load(15);
   u6[2];
   u7;
   u8;
   return float4(u8.Consume() + t2.SampleCmp(s2, 1.0, 2.0)) + t1.Sample(s1, 1.0)
    + t6.Load(1);
}
"""

@inside_glslc_testsuite('OptionFAutoBindUniforms')
class UniformBindingsNotCreatedByDefault(expect.ErrorMessageSubstr):
    """Tests that compilation fails when uniforms have no binding."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader]
    expected_error_substr = "sampler/texture/image requires layout(binding=X)"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FAutoBindingUniformsGeneratesBindings(expect.ValidAssemblyFileWithSubstr):
    """Tests that the compiler generates bindings for uniforms upon request ."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms']
    # Sufficient to just check one of the uniforms.
    expected_assembly_substr = "OpDecorate %my_sam Binding 1"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FImageBindingBaseOptionRespectedOnImage(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fimage-binding-base value is respected on images."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-fimage-binding-base', '44']
    expected_assembly_substr = "OpDecorate %my_img Binding 44"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FImageBindingBaseOptionRespectedOnImageBuffer(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fimage-binding-base value is respected on image buffers."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-fimage-binding-base', '44']
    expected_assembly_substr = "OpDecorate %my_imbuf Binding 45"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FTextureBindingBaseOptionRespected(expect.ValidAssemblyFileWithSubstr):
    """Tests that -ftexture-binding-base value is respected."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-ftexture-binding-base', '44']
    expected_assembly_substr = "OpDecorate %my_tex Binding 44"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FSamplerBindingBaseOptionRespected(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fsampler-binding-base value is respected."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-fsampler-binding-base', '44']
    expected_assembly_substr = "OpDecorate %my_sam Binding 44"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FUboBindingBaseOptionRespectedOnBuffer(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fubo-binding-base value is respected."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-fubo-binding-base', '44']
    expected_assembly_substr = "OpDecorate %my_ubo Binding 44"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FCbufferBindingBaseOptionRespectedOnBuffer(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fcbuffer-binding-base value is respected."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-fcbuffer-binding-base', '44']
    expected_assembly_substr = "OpDecorate %my_ubo Binding 44"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FImageBindingBaseNeedsValue(expect.ErrorMessageSubstr):
    """Tests that -fimage-binding-base requires a value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fimage-binding-base']
    expected_error_substr = "error: Option -fimage-binding-base requires at least one argument"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FTextureBindingBaseNeedsValue(expect.ErrorMessageSubstr):
    """Tests that -ftexture-binding-base requires a value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-ftexture-binding-base']
    expected_error_substr = "error: Option -ftexture-binding-base requires at least one argument"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FSamplerBindingBaseNeedsValue(expect.ErrorMessageSubstr):
    """Tests that -fsampler-binding-base requires a value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fsampler-binding-base']
    expected_error_substr = "error: Option -fsampler-binding-base requires at least one argument"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FUboBindingBaseNeedsValue(expect.ErrorMessageSubstr):
    """Tests that -fubo-binding-base requires a value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fubo-binding-base']
    expected_error_substr = "error: Option -fubo-binding-base requires at least one argument"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FImageBindingBaseNeedsNumberValueIfNotStage(expect.ErrorMessageSubstr):
    """Tests that -fimage-binding-base requires a number value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fimage-binding-base', '9x']
    expected_error_substr = "error: invalid offset value 9x for -fimage-binding-base"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FTextureBindingBaseNeedsNumberValue(expect.ErrorMessageSubstr):
    """Tests that -ftexture-binding-base requires a number value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-ftexture-binding-base', '9x']
    expected_error_substr = "error: invalid offset value 9x for -ftexture-binding-base"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FSamplerBindingBaseNeedsNumberValue(expect.ErrorMessageSubstr):
    """Tests that -fsampler-binding-base requires a number value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fsampler-binding-base', '9x']
    expected_error_substr = "error: invalid offset value 9x for -fsampler-binding-base"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FUboBindingBaseNeedsNumberValue(expect.ErrorMessageSubstr):
    """Tests that -fubo-binding-base requires a number value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fubo-binding-base', '9x']
    expected_error_substr = "error: invalid offset value 9x for -fubo-binding-base"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FImageBindingBaseNeedsUnsignedNumberValue(expect.ErrorMessageSubstr):
    """Tests that -fimage-binding-base requires an unsigned number value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fimage-binding-base', '-6']
    expected_error_substr = "error: invalid offset value -6 for -fimage-binding-base"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FTextureBindingBaseNeedsUnsignedNumberValue(expect.ErrorMessageSubstr):
    """Tests that -ftexture-binding-base requires an unsigned number value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-ftexture-binding-base', '-6']
    expected_error_substr = "error: invalid offset value -6 for -ftexture-binding-base"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FSamplerBindingBaseNeedsUnsignedNumberValue(expect.ErrorMessageSubstr):
    """Tests that -fsampler-binding-base requires an unsigned value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fsampler-binding-base', '-6']
    expected_error_substr = "error: invalid offset value -6 for -fsampler-binding-base"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FUboBindingBaseNeedsUnsignedNumberValue(expect.ErrorMessageSubstr):
    """Tests that -fubo-binding-base requires an unsigned value."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fubo-binding-base', '-6']
    expected_error_substr = "error: invalid offset value -6 for -fubo-binding-base"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FImageBindingBaseForVertOptionRespectedOnImageCompileAsVert(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fimage-binding-base with vert stage value is respected on images."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-fimage-binding-base', 'vert', '44']
    expected_assembly_substr = "OpDecorate %my_img Binding 44"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FImageBindingBaseForVertOptionIgnoredOnImageCompileAsFrag(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fimage-binding-base with vert stage value is ignored when cmopiled as
    fragment."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-fimage-binding-base', 'vert', '44']
    expected_assembly_substr = "OpDecorate %my_img Binding 2"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FImageBindingBaseForFragOptionRespectedOnImageCompileAsFrag(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fimage-binding-base with frag stage value is respected on images."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-fimage-binding-base', 'frag', '44']
    expected_assembly_substr = "OpDecorate %my_img Binding 44"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FSsboBindingBaseRespectedOnSsboCompileAsFrag(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fssbo-binding-base with frag stage value is respected on SSBOs."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-fssbo-binding-base', '100']
    expected_assembly_substr = "OpDecorate %my_ssbo Binding 100"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FSsboBindingBaseForFragOptionRespectedOnSsboCompileAsFrag(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fssbo-binding-base with frag stage value is respected on SSBOs."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-fssbo-binding-base', 'frag', '100']
    expected_assembly_substr = "OpDecorate %my_ssbo Binding 100"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class FSsboBindingBaseForFragOptionIgnoredOnSsboCompileAsVert(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fssbo-binding-base with frag stage value is ignored on SSBOs
    when compiling as a vertex shader."""

    shader = FileShader(GLSL_SHADER_WITH_UNIFORMS_WITHOUT_BINDINGS, '.vert')
    glslc_args = ['-S', shader, '-fauto-bind-uniforms', '-fssbo-binding-base', 'frag', '100']
    expected_assembly_substr = "OpDecorate %my_ssbo Binding 5"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class AutomaticHlslIoMapping(expect.ValidAssemblyFileWithSubstr):
    """Tests that HLSL IO Mapping uses the register values in the source code."""

    shader = FileShader(HLSL_SHADER_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl-iomap',
                  '-fauto-bind-uniforms', shader]
    expected_assembly_substr = "OpDecorate %u8 Binding 28"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class HlslFSamplerBindingBaseOptionRespected(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fsampler-binding-base sets binding base for samplers in HLSL
    compilation."""

    shader = FileShader(HLSL_SHADER_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl-iomap', shader,
                  '-fauto-bind-uniforms', '-fsampler-binding-base', '100']
    expected_assembly_substr = "OpDecorate %s2 Binding 102"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class HlslFSamplerBindingBaseForFragOptionRespected(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fsampler-binding-base for frag sets binding base for samplers
    in HLSL compilation."""

    shader = FileShader(HLSL_SHADER_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl-iomap', shader,
                  '-fauto-bind-uniforms', '-fsampler-binding-base', 'frag', '100']
    expected_assembly_substr = "OpDecorate %s2 Binding 102"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class HlslFSamplerBindingBaseForComputeOptionIgnoredWhenCompilingAsFrag(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fsampler-binding-base for compute is ignored when compiling
    as a fragment shader."""

    shader = FileShader(HLSL_SHADER_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl-iomap', shader,
                  '-fauto-bind-uniforms', '-fsampler-binding-base', 'compute', '100']
    expected_assembly_substr = "OpDecorate %s2 Binding 2"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class HlslFTextureBindingBaseOptionRespected(expect.ValidAssemblyFileWithSubstr):
    """Tests that -ftexture-binding-base sets binding base for textures in HLSL
    compilation."""

    shader = FileShader(HLSL_SHADER_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl-iomap', shader,
                  '-fauto-bind-uniforms', '-ftexture-binding-base', '100']
    expected_assembly_substr = "OpDecorate %t6 Binding 116"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class HlslFTextureBindingBaseForFragOptionRespected(expect.ValidAssemblyFileWithSubstr):
    """Tests that -ftexture-binding-base for frag sets binding base for textures
    in HLSL compilation."""

    shader = FileShader(HLSL_SHADER_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl-iomap', shader,
                  '-fauto-bind-uniforms', '-ftexture-binding-base', 'frag', '100']
    expected_assembly_substr = "OpDecorate %t6 Binding 116"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class HlslFTextureBindingBaseForComputeOptionIgnoredWhenCompilingAsFrag(expect.ValidAssemblyFileWithSubstr):
    """Tests that -ftexture-binding-base for compute is ignored when compiling
    as a fragment shader."""

    shader = FileShader(HLSL_SHADER_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl-iomap', shader,
                  '-fauto-bind-uniforms', '-ftexture-binding-base', 'compute', '100']
    expected_assembly_substr = "OpDecorate %t6 Binding 16"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class HlslFUavBindingBaseOptionRespected(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fuav-binding-base sets binding base for UAVs in HLSL
    compilation."""

    shader = FileShader(HLSL_SHADER_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl-iomap', shader,
                  '-fauto-bind-uniforms', '-fuav-binding-base', '100']
    expected_assembly_substr = "OpDecorate %u8 Binding 128"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class HlslFUavBindingBaseForFragOptionRespected(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fuav-binding-base for frag sets binding base for UAVs in HLSL
    compilation."""

    shader = FileShader(HLSL_SHADER_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl-iomap', shader,
                  '-fauto-bind-uniforms', '-fuav-binding-base', 'frag', '100']
    expected_assembly_substr = "OpDecorate %u8 Binding 128"


@inside_glslc_testsuite('OptionFAutoBindUniforms')
class HlslFUavBindingBaseForComputeOptionIgnoredWhenCompilingAsFrag(expect.ValidAssemblyFileWithSubstr):
    """Tests that -fuav-binding-base for compute is ignored when compiling
    as a fragment shader."""

    shader = FileShader(HLSL_SHADER_WITHOUT_BINDINGS, '.frag')
    glslc_args = ['-S', '-x', 'hlsl', '-fhlsl-iomap', shader,
                  '-fauto-bind-uniforms', '-fuav-binding-base', 'compute', '100']
    expected_assembly_substr = "OpDecorate %u8 Binding 28"



# Copyright 2023 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader

# A GLSL shader with unused bindings.
GLSL_SHADER_WITH_UNUSED_BINDINGS = """#version 450
  layout(binding=0) buffer InputA { vec4 x[]; } inputA;
  layout(binding=1) buffer InputB { vec4 x[]; } inputB;
  layout(binding=2) buffer Output { vec4 x[]; } outputO;

  void main() {}
  """


@inside_glslc_testsuite('OptionFPreserveBindings')
class TestFPreserveBindingsInputA(expect.ValidAssemblyFileWithSubstr):
    """Tests that the compiler preserves bindings when optimizations are
    enabled."""

    shader = FileShader(GLSL_SHADER_WITH_UNUSED_BINDINGS, '.comp')
    glslc_args = ['-S', '-O', shader, '-fpreserve-bindings']
    # Check the first buffer.
    expected_assembly_substr = "Binding 0"


@inside_glslc_testsuite('OptionFPreserveBindings')
class TestFPreserveBindingsInputB(expect.ValidAssemblyFileWithSubstr):
    """Tests that the compiler preserves bindings when optimizations are
    enabled."""

    shader = FileShader(GLSL_SHADER_WITH_UNUSED_BINDINGS, '.comp')
    glslc_args = ['-S', '-O', shader, '-fpreserve-bindings']
    # Check the first buffer.
    expected_assembly_substr = "Binding 1"


@inside_glslc_testsuite('OptionFPreserveBindings')
class TestFPreserveBindingsOutputO(expect.ValidAssemblyFileWithSubstr):
    """Tests that the compiler preserves bindings when optimizations are
    enabled."""

    shader = FileShader(GLSL_SHADER_WITH_UNUSED_BINDINGS, '.comp')
    glslc_args = ['-S', '-O', shader, '-fpreserve-bindings']
    # Check the first buffer.
    expected_assembly_substr = "Binding 2"


@inside_glslc_testsuite('OptionFPreserveBindings')
class TestFNoPreserveBindings(expect.ValidAssemblyFileWithoutSubstr):
    """Tests that the compiler removes bindings when -fpreserve-bindings is not
    set."""

    shader = FileShader(GLSL_SHADER_WITH_UNUSED_BINDINGS, '.comp')
    glslc_args = ['-S', '-O', shader]
    # Check that all binding decorations are gone.
    unexpected_assembly_substr = "OpDecorate"



# Copyright 2015 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A number of common glslc result checks coded in mixin classes.

A test case can use these checks by declaring their enclosing mixin classes
as superclass and providing the expected_* variables required by the check_*()
methods in the mixin classes.
"""
import difflib
import functools
import os
import re
import subprocess
import sys
from glslc_test_framework import GlslCTest
from builtins import bytes

GLSLANG_GENERATOR_VERSION=11
SHADERC_GENERATOR_NUMBER=13
SHADERC_GENERATOR_WORD=(SHADERC_GENERATOR_NUMBER << 16) + GLSLANG_GENERATOR_VERSION
ASSEMBLER_GENERATOR_WORD=(7<<16)

def convert_to_string(input):
    if type(input) is not str:
        if sys.version_info[0] == 2:
            return input.decode('utf-8')
        elif sys.version_info[0] == 3:
            return str(input,
                              encoding='utf-8',
                              errors='ignore') if input is not None else input
        else:
            raise Exception(
                'Unable to determine if running Python 2 or 3 from {}'.format(
                    sys.version_info))
    else:
        return input


def convert_to_unix_line_endings(source):
    """Converts all line endings in source to be unix line endings."""
    return source.replace('\r\n', '\n').replace('\r', '\n')


def substitute_file_extension(filename, extension):
    """Substitutes file extension, respecting known shader extensions.

    foo.vert -> foo.vert.[extension] [similarly for .frag, .comp, etc.]
    foo.glsl -> foo.[extension]
    foo.unknown -> foo.[extension]
    foo -> foo.[extension]
    """
    if filename[-5:] not in ['.vert', '.frag', '.tesc', '.tese',
                             '.geom', '.comp', '.spvasm']:
        return filename.rsplit('.', 1)[0] + '.' + extension
    else:
        return filename + '.' + extension


def get_object_filename(source_filename):
    """Gets the object filename for the given source file."""
    return substitute_file_extension(source_filename, 'spv')


def get_assembly_filename(source_filename):
    """Gets the assembly filename for the given source file."""
    return substitute_file_extension(source_filename, 'spvasm')


def verify_file_non_empty(filename):
    """Checks that a given file exists and is not empty."""
    if not os.path.isfile(filename):
        return False, 'Cannot find file: ' + filename
    if not os.path.getsize(filename):
        return False, 'Empty file: ' + filename
    return True, ''


class ReturnCodeIsZero(GlslCTest):
    """Mixin class for checking that the return code is zero."""

    def check_return_code_is_zero(self, status):
        if status.returncode:
            return False, 'Non-zero return code: {ret}\n'.format(
                ret=status.returncode)
        return True, ''


class NoOutputOnStdout(GlslCTest):
    """Mixin class for checking that there is no output on stdout."""

    def check_no_output_on_stdout(self, status):
        if status.stdout:
            return False, 'Non empty stdout: {out}\n'.format(out=status.stdout)
        return True, ''


class NoOutputOnStderr(GlslCTest):
    """Mixin class for checking that there is no output on stderr."""

    def check_no_output_on_stderr(self, status):
        if status.stderr:
            return False, 'Non empty stderr: {err}\n'.format(err=status.stderr)
        return True, ''


class SuccessfulReturn(ReturnCodeIsZero, NoOutputOnStdout, NoOutputOnStderr):
    """Mixin class for checking that return code is zero and no output on
    stdout and stderr."""
    pass


class NoGeneratedFiles(GlslCTest):
    """Mixin class for checking that there is no file generated."""

    def check_no_generated_files(self, status):
        all_files = os.listdir(status.directory)
        input_files = status.input_filenames
        if all([f.startswith(status.directory) for f in input_files]):
            all_files = [os.path.join(status.directory, f) for f in all_files]
        generated_files = set(all_files) - set(input_files)
        if len(generated_files) == 0:
            return True, ''
        else:
            return False, 'Extra files generated: {}'.format(generated_files)


class CorrectBinaryLengthAndPreamble(GlslCTest):
    """Provides methods for verifying preamble for a SPIR-V binary."""

    def verify_binary_length_and_header(self, binary, spv_version = 0x10000):
        """Checks that the given SPIR-V binary has valid length and header.

        Returns:
            False, error string if anything is invalid
            True, '' otherwise
        Args:
            binary: a bytes object containing the SPIR-V binary
            spv_version: target SPIR-V version number, with same encoding
                 as the version word in a SPIR-V header.
        """

        def read_word(binary, index, little_endian):
            """Reads the index-th word from the given binary file."""
            word = binary[index * 4:(index + 1) * 4]
            if little_endian:
                word = reversed(word)
            return functools.reduce(lambda w, b: (w << 8) | b, word, 0)

        def check_endianness(binary):
            """Checks the endianness of the given SPIR-V binary.

            Returns:
              True if it's little endian, False if it's big endian.
              None if magic number is wrong.
            """
            first_word = read_word(binary, 0, True)
            if first_word == 0x07230203:
                return True
            first_word = read_word(binary, 0, False)
            if first_word == 0x07230203:
                return False
            return None

        num_bytes = len(binary)
        if num_bytes % 4 != 0:
            return False, ('Incorrect SPV binary: size should be a multiple'
                           ' of words')
        if num_bytes < 20:
            return False, 'Incorrect SPV binary: size less than 5 words'

        preamble = binary[0:19]
        little_endian = check_endianness(preamble)
        # SPIR-V module magic number
        if little_endian is None:
            return False, 'Incorrect SPV binary: wrong magic number'

        # SPIR-V version number
        version = read_word(preamble, 1, little_endian)
        # TODO(dneto): Recent Glslang uses version word 0 for opengl_compat
        # profile

        if version != spv_version and version != 0:
            return False, 'Incorrect SPV binary: wrong version number'
        # Shaderc-over-Glslang (0x000d....) or
        # SPIRV-Tools (0x0007....) generator number
        if read_word(preamble, 2, little_endian) != SHADERC_GENERATOR_WORD and \
                read_word(preamble, 2, little_endian) != ASSEMBLER_GENERATOR_WORD:
            return False, ('Incorrect SPV binary: wrong generator magic '
                           'number')
        # reserved for instruction schema
        if read_word(preamble, 4, little_endian) != 0:
            return False, 'Incorrect SPV binary: the 5th byte should be 0'

        return True, ''


class CorrectObjectFilePreamble(CorrectBinaryLengthAndPreamble):
    """Provides methods for verifying preamble for a SPV object file."""

    def verify_object_file_preamble(self, filename, spv_version = 0x10000):
        """Checks that the given SPIR-V binary file has correct preamble."""

        success, message = verify_file_non_empty(filename)
        if not success:
            return False, message

        with open(filename, 'rb') as object_file:
            object_file.seek(0, os.SEEK_END)
            num_bytes = object_file.tell()

            object_file.seek(0)

            binary = bytes(object_file.read())
            return self.verify_binary_length_and_header(binary, spv_version)

        return True, ''


class CorrectAssemblyFilePreamble(GlslCTest):
    """Provides methods for verifying preamble for a SPV assembly file."""

    def verify_assembly_file_preamble(self, filename):
        success, message = verify_file_non_empty(filename)
        if not success:
            return False, message

        with open(filename) as assembly_file:
            line1 = assembly_file.readline()
            line2 = assembly_file.readline()
            line3 = assembly_file.readline()

        if (line1 != '; SPIR-V\n' or
            line2 != '; Version: 1.0\n' or
            (not line3.startswith('; Generator: Google Shaderc over Glslang;'))):
            return False, 'Incorrect SPV assembly'

        return True, ''


class ValidObjectFile(SuccessfulReturn, CorrectObjectFilePreamble):
    """Mixin class for checking that every input file generates a valid SPIR-V 1.0
    object file following the object file naming rule, and there is no output on
    stdout/stderr."""

    def check_object_file_preamble(self, status):
        for input_filename in status.input_filenames:
            object_filename = get_object_filename(input_filename)
            success, message = self.verify_object_file_preamble(
                os.path.join(status.directory, object_filename))
            if not success:
                return False, message
        return True, ''


class ValidObjectFile1_3(SuccessfulReturn, CorrectObjectFilePreamble):
    """Mixin class for checking that every input file generates a valid SPIR-V 1.3
    object file following the object file naming rule, and there is no output on
    stdout/stderr."""

    def check_object_file_preamble(self, status):
        for input_filename in status.input_filenames:
            object_filename = get_object_filename(input_filename)
            success, message = self.verify_object_file_preamble(
                os.path.join(status.directory, object_filename),
                0x10300)
            if not success:
                return False, message
        return True, ''


class ValidObjectFile1_4(SuccessfulReturn, CorrectObjectFilePreamble):
    """Mixin class for checking that every input file generates a valid SPIR-V 1.4
    object file following the object file naming rule, and there is no output on
    stdout/stderr."""

    def check_object_file_preamble(self, status):
        for input_filename in status.input_filenames:
            object_filename = get_object_filename(input_filename)
            success, message = self.verify_object_file_preamble(
                os.path.join(status.directory, object_filename),
                0x10400)
            if not success:
                return False, message
        return True, ''


class ValidObjectFile1_5(SuccessfulReturn, CorrectObjectFilePreamble):
    """Mixin class for checking that every input file generates a valid SPIR-V 1.5
    object file following the object file naming rule, and there is no output on
    stdout/stderr."""

    def check_object_file_preamble(self, status):
        for input_filename in status.input_filenames:
            object_filename = get_object_filename(input_filename)
            success, message = self.verify_object_file_preamble(
                os.path.join(status.directory, object_filename),
                0x10500)
            if not success:
                return False, message
        return True, ''


class ValidObjectFile1_6(SuccessfulReturn, CorrectObjectFilePreamble):
    """Mixin class for checking that every input file generates a valid SPIR-V 1.6
    object file following the object file naming rule, and there is no output on
    stdout/stderr."""

    def check_object_file_preamble(self, status):
        for input_filename in status.input_filenames:
            object_filename = get_object_filename(input_filename)
            success, message = self.verify_object_file_preamble(
                os.path.join(status.directory, object_filename),
                0x10600)
            if not success:
                return False, message
        return True, ''


class ValidObjectFileWithAssemblySubstr(SuccessfulReturn, CorrectObjectFilePreamble):
    """Mixin class for checking that every input file generates a valid object
    file following the object file naming rule, there is no output on
    stdout/stderr, and the disassmbly contains a specified substring per input."""

    def check_object_file_disassembly(self, status):
        for an_input in status.inputs:
            object_filename = get_object_filename(an_input.filename)
            obj_file = str(os.path.join(status.directory, object_filename))
            success, message = self.verify_object_file_preamble(obj_file)
            if not success:
                return False, message
            cmd = [status.test_manager.disassembler_path, '--no-color', obj_file]
            process = subprocess.Popen(
                args=cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE, cwd=status.directory)
            output = process.communicate(None)
            disassembly = output[0]
            if not isinstance(an_input.assembly_substr, str):
                return False, "Missing assembly_substr member"
            if bytes(an_input.assembly_substr, 'utf-8') not in disassembly:
                return False, ('Incorrect disassembly output:\n{asm}\n'
                    'Expected substring not found:\n{exp}'.format(
                    asm=disassembly, exp=an_input.assembly_substr))
        return True, ''


class ValidNamedObjectFile(SuccessfulReturn, CorrectObjectFilePreamble):
    """Mixin class for checking that a list of object files with the given
    names are correctly generated, and there is no output on stdout/stderr.

    To mix in this class, subclasses need to provide expected_object_filenames
    as the expected object filenames.
    """

    def check_object_file_preamble(self, status):
        for object_filename in self.expected_object_filenames:
            success, message = self.verify_object_file_preamble(
                os.path.join(status.directory, object_filename))
            if not success:
                return False, message
        return True, ''


class ValidFileContents(GlslCTest):
    """Mixin class to test that a specific file contains specific text
    To mix in this class, subclasses need to provide expected_file_contents as
    the contents of the file and target_filename to determine the location."""

    def check_file(self, status):
        target_filename = os.path.join(status.directory, self.target_filename)
        if not os.path.isfile(target_filename):
            return False, 'Cannot find file: ' + target_filename
        with open(target_filename, 'r') as target_file:
            file_contents = target_file.read()
            if isinstance(self.expected_file_contents, str):
                if file_contents == self.expected_file_contents:
                    return True, ''
                return False, ('Incorrect file output: \n{act}\n'
                               'Expected:\n{exp}'
                               'With diff:\n{diff}'.format(
                                   act=file_contents,
                                   exp=self.expected_file_contents,
                                   diff='\n'.join(list(difflib.unified_diff(
                                       self.expected_file_contents.split('\n'),
                                       file_contents.split('\n'),
                                       fromfile='expected_output',
                                       tofile='actual_output')))))
            elif isinstance(self.expected_file_contents, type(re.compile(''))):
                if self.expected_file_contents.search(file_contents):
                    return True, ''
                return False, (
                    'Incorrect file output: \n{act}\n'
                    'Expected matching regex pattern:\n{exp}'.format(
                        act=file_contents,
                        exp=self.expected_file_contents.pattern))
        return False, ('Could not open target file ' + target_filename +
                       ' for reading')


class ValidAssemblyFile(SuccessfulReturn, CorrectAssemblyFilePreamble):
    """Mixin class for checking that every input file generates a valid assembly
    file following the assembly file naming rule, and there is no output on
    stdout/stderr."""

    def check_assembly_file_preamble(self, status):
        for input_filename in status.input_filenames:
            assembly_filename = get_assembly_filename(input_filename)
            success, message = self.verify_assembly_file_preamble(
                os.path.join(status.directory, assembly_filename))
            if not success:
                return False, message
        return True, ''


class ValidAssemblyFileWithSubstr(ValidAssemblyFile):
    """Mixin class for checking that every input file generates a valid assembly
    file following the assembly file naming rule, there is no output on
    stdout/stderr, and all assembly files have the given substring specified
    by expected_assembly_substr.

    To mix in this class, subclasses need to provde expected_assembly_substr
    as the expected substring.
    """

    def check_assembly_with_substr(self, status):
        for input_filename in status.input_filenames:
            assembly_filename = get_assembly_filename(input_filename)
            success, message = self.verify_assembly_file_preamble(
                os.path.join(status.directory, assembly_filename))
            if not success:
                return False, message
            with open(assembly_filename, 'r') as f:
                content = f.read()
                if self.expected_assembly_substr not in convert_to_unix_line_endings(content):
                   return False, ('Incorrect assembly output:\n{asm}\n'
                                  'Expected substring not found:\n{exp}'.format(
                                  asm=content, exp=self.expected_assembly_substr))
        return True, ''


class ValidAssemblyFileWithoutSubstr(ValidAssemblyFile):
    """Mixin class for checking that every input file generates a valid assembly
    file following the assembly file naming rule, there is no output on
    stdout/stderr, and no assembly files have the given substring specified
    by unexpected_assembly_substr.

    To mix in this class, subclasses need to provde unexpected_assembly_substr
    as the substring we expect not to see.
    """

    def check_assembly_for_substr(self, status):
        for input_filename in status.input_filenames:
            assembly_filename = get_assembly_filename(input_filename)
            success, message = self.verify_assembly_file_preamble(
                os.path.join(status.directory, assembly_filename))
            if not success:
                return False, message
            with open(assembly_filename, 'r') as f:
                content = f.read()
                if self.unexpected_assembly_substr in convert_to_unix_line_endings(content):
                   return False, ('Incorrect assembly output:\n{asm}\n'
                                  'Unexpected substring found:\n{unexp}'.format(
                                  asm=content, exp=self.unexpected_assembly_substr))
        return True, ''


class ValidNamedAssemblyFile(SuccessfulReturn, CorrectAssemblyFilePreamble):
    """Mixin class for checking that a list of assembly files with the given
    names are correctly generated, and there is no output on stdout/stderr.

    To mix in this class, subclasses need to provide expected_assembly_filenames
    as the expected assembly filenames.
    """

    def check_object_file_preamble(self, status):
        for assembly_filename in self.expected_assembly_filenames:
            success, message = self.verify_assembly_file_preamble(
                os.path.join(status.directory, assembly_filename))
            if not success:
                return False, message
        return True, ''


class ErrorMessage(GlslCTest):
    """Mixin class for tests that fail with a specific error message.

    To mix in this class, subclasses need to provide expected_error as the
    expected error message.

    The test should fail if the subprocess was terminated by a signal.
    """

    def check_has_error_message(self, status):
        if not status.returncode:
            return False, ('Expected error message, but returned success from '
                           'glslc')
        if status.returncode < 0:
            # On Unix, a negative value -N for Popen.returncode indicates
            # termination by signal N.
            # https://docs.python.org/2/library/subprocess.html
            return False, ('Expected error message, but glslc was terminated by '
                           'signal ' + str(status.returncode))
        if not status.stderr:
            return False, 'Expected error message, but no output on stderr'
        if self.expected_error != convert_to_unix_line_endings(convert_to_string(status.stderr)):
            return False, ('Incorrect stderr output:\n{act}\n'
                           'Expected:\n{exp}'.format(
                               act=status.stderr, exp=self.expected_error))
        return True, ''


class ErrorMessageSubstr(GlslCTest):
    """Mixin class for tests that fail with a specific substring in the error
    message.

    To mix in this class, subclasses need to provide expected_error_substr as
    the expected error message substring.

    The test should fail if the subprocess was terminated by a signal.
    """

    def check_has_error_message_as_substring(self, status):
        if not status.returncode:
            return False, ('Expected error message, but returned success from '
                           'glslc')
        if status.returncode < 0:
            # On Unix, a negative value -N for Popen.returncode indicates
            # termination by signal N.
            # https://docs.python.org/2/library/subprocess.html
            return False, ('Expected error message, but glslc was terminated by '
                           'signal ' + str(status.returncode))
        if not status.stderr:
            return False, 'Expected error message, but no output on stderr'
        if self.expected_error_substr not in convert_to_unix_line_endings(convert_to_string(status.stderr)):
            return False, ('Incorrect stderr output:\n{act}\n'
                           'Expected substring not found in stderr:\n{exp}'.format(
                               act=status.stderr, exp=self.expected_error_substr))
        return True, ''


class WarningMessage(GlslCTest):
    """Mixin class for tests that succeed but have a specific warning message.

    To mix in this class, subclasses need to provide expected_warning as the
    expected warning message.
    """

    def check_has_warning_message(self, status):
        if status.returncode:
            return False, ('Expected warning message, but returned failure from'
                           ' glslc')
        if not status.stderr:
            return False, 'Expected warning message, but no output on stderr'
        if self.expected_warning != convert_to_unix_line_endings(convert_to_string(status.stderr)):
            return False, ('Incorrect stderr output:\n{act}\n'
                           'Expected:\n{exp}'.format(
                               act=status.stderr, exp=self.expected_warning))
        return True, ''


class ValidObjectFileWithWarning(
    NoOutputOnStdout, CorrectObjectFilePreamble, WarningMessage):
    """Mixin class for checking that every input file generates a valid object
    file following the object file naming rule, with a specific warning message.
    """

    def check_object_file_preamble(self, status):
        for input_filename in status.input_filenames:
            object_filename = get_object_filename(input_filename)
            success, message = self.verify_object_file_preamble(
                os.path.join(status.directory, object_filename))
            if not success:
                return False, message
        return True, ''


class ValidAssemblyFileWithWarning(
    NoOutputOnStdout, CorrectAssemblyFilePreamble, WarningMessage):
    """Mixin class for checking that every input file generates a valid assembly
    file following the assembly file naming rule, with a specific warning
    message."""

    def check_assembly_file_preamble(self, status):
        for input_filename in status.input_filenames:
            assembly_filename = get_assembly_filename(input_filename)
            success, message = self.verify_assembly_file_preamble(
                os.path.join(status.directory, assembly_filename))
            if not success:
                return False, message
        return True, ''


class StdoutMatch(GlslCTest):
    """Mixin class for tests that can expect output on stdout.

    To mix in this class, subclasses need to provide expected_stdout as the
    expected stdout output.

    For expected_stdout, if it's True, then they expect something on stdout but
    will not check what it is. If it's a string, expect an exact match.  If it's
    anything else, expect expected_stdout.search(stdout) to be true.
    """

    def check_stdout_match(self, status):
        # "True" in this case means we expect something on stdout, but we do not
        # care what it is, we want to distinguish this from "blah" which means we
        # expect exactly the string "blah".
        if self.expected_stdout is True:
            if not status.stdout:
                return False, 'Expected something on stdout'
        elif type(self.expected_stdout) == str:
            if self.expected_stdout != convert_to_unix_line_endings(
                    convert_to_string(status.stdout)):
                return False, ('Incorrect stdout output:\n{ac}\n'
                               'Expected:\n{ex}'.format(
                                   ac=status.stdout, ex=self.expected_stdout))
        else:
            if not self.expected_stdout.search(convert_to_unix_line_endings(
                    convert_to_string(status.stdout))):
                return False, ('Incorrect stdout output:\n{ac}\n'
                               'Expected to match regex:\n{ex}'.format(
                                   ac=convert_to_string(status.stdout),
                                   ex=self.expected_stdout.pattern))
        return True, ''


class StderrMatch(GlslCTest):
    """Mixin class for tests that can expect output on stderr.

    To mix in this class, subclasses need to provide expected_stderr as the
    expected stderr output.

    For expected_stderr, if it's True, then they expect something on stderr,
    but will not check what it is. If it's a string, expect an exact match.
    """

    def check_stderr_match(self, status):
        # "True" in this case means we expect something on stderr, but we do not
        # care what it is, we want to distinguish this from "blah" which means we
        # expect exactly the string "blah".
        if self.expected_stderr is True:
            if not status.stderr:
                return False, 'Expected something on stderr'
        else:
            if self.expected_stderr != convert_to_unix_line_endings(
                    convert_to_string(status.stderr)):
                return False, ('Incorrect stderr output:\n{ac}\n'
                               'Expected:\n{ex}'.format(
                                   ac=status.stderr, ex=self.expected_stderr))
        return True, ''


class StdoutNoWiderThan80Columns(GlslCTest):
    """Mixin class for tests that require stdout to 80 characters or narrower.

    To mix in this class, subclasses need to provide expected_stdout as the
    expected stdout output.
    """

    def check_stdout_not_too_wide(self, status):
        if not status.stdout:
            return True, ''
        else:
            for line in status.stdout.splitlines():
                if len(line) > 80:
                    return False, ('Stdout line longer than 80 columns: %s'
                                   % line)
        return True, ''


class NoObjectFile(GlslCTest):
    """Mixin class for checking that no input file has a corresponding object
    file."""

    def check_no_object_file(self, status):
        for input_filename in status.input_filenames:
            object_filename = get_object_filename(input_filename)
            full_object_file = os.path.join(status.directory, object_filename)
            print("checking %s" % full_object_file)
            if os.path.isfile(full_object_file):
                return False, ('Expected no object file, but found: %s'
                               % full_object_file)
        return True, ''


class NoNamedOutputFiles(GlslCTest):
    """Mixin class for checking that no specified output files exist.

    The expected_output_filenames member should be full pathnames."""

    def check_no_named_output_files(self, status):
        for object_filename in self.expected_output_filenames:
            if os.path.isfile(object_filename):
                return False, ('Expected no output file, but found: %s'
                               % object_filename)
        return True, ''



# Copyright 2016 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
from environment import File, Directory
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader


def shader_source_with_tex_offset(offset):
    """Returns a vertex shader using a texture access with the given offset."""

    return """#version 450
              layout (binding=0) uniform sampler1D tex;
              void main() { vec4 x = textureOffset(tex, 1.0, """ + str(offset) + "); }"


def shader_with_tex_offset(offset):
    """Returns a vertex FileShader using a texture access with the given offset."""

    return FileShader(shader_source_with_tex_offset(offset), ".vert")

@inside_glslc_testsuite('OptionFLimit')
class TestFLimitNoEqual(expect.ErrorMessage):
    """Tests -flimit without equal."""

    glslc_args = ['-flimit']
    expected_error = ["glslc: error: unknown argument: '-flimit'\n"]


@inside_glslc_testsuite('OptionFLimit')
class TestFLimitJustEqual(expect.ValidObjectFile):
    """Tests -flimit= with no argument."""

    shader = shader_with_tex_offset(0);
    glslc_args = ['-c', shader, '-flimit=']


@inside_glslc_testsuite('OptionFLimit')
class TestFLimitJustEqualMaxOffset(expect.ValidObjectFile):
    """Tests -flimit= with no argument.  The shader uses max offset."""

    shader = shader_with_tex_offset(7);
    glslc_args = ['-c', shader, '-flimit=']


@inside_glslc_testsuite('OptionFLimit')
class TestFLimitJustEqualMinOffset(expect.ValidObjectFile):
    """Tests -flimit= with no argument.  The shader uses min offset."""

    shader = shader_with_tex_offset(-8);
    glslc_args = ['-c', shader, '-flimit=']


@inside_glslc_testsuite('OptionFLimit')
class TestFLimitJustEqualBelowMinOffset(expect.ErrorMessageSubstr):
    """Tests -flimit= with no argument.  The shader uses below min default offset."""

    shader = shader_with_tex_offset(-9);
    glslc_args = ['-c', shader, '-flimit=']
    expected_error_substr = ["'texel offset' : value is out of range"]


@inside_glslc_testsuite('OptionFLimit')
class TestFLimitLowerThanDefaultMinOffset(expect.ValidObjectFile):
    """Tests -flimit= with lower than default argument.  The shader uses below min offset."""

    shader = shader_with_tex_offset(-9);
    glslc_args = ['-c', shader, '-flimit= MinProgramTexelOffset -9 ']


@inside_glslc_testsuite('OptionFLimit')
class TestFLimitIgnoredLangFeatureSettingSample(expect.ValidObjectFile):
    """Tests -flimit= an ignored option."""

    shader = FileShader("#version 150\nvoid main() { while(true); }", '.vert')
    glslc_args = ['-c', shader, '-flimit=whileLoops 0']


@inside_glslc_testsuite('OptionFLimit')
class TestFLimitLowerThanDefaultMinOffset(expect.ValidObjectFile):
    """Tests -flimit= with lower than default argument.  The shader uses that offset."""

    shader = shader_with_tex_offset(-9);
    glslc_args = ['-c', shader, '-flimit= MinProgramTexelOffset -9 ']


@inside_glslc_testsuite('OptionFLimitFile')
class TestFLimitFileNoArg(expect.ErrorMessage):
    """Tests -flimit-file without an argument"""

    shader = shader_with_tex_offset(-9);
    glslc_args = ['-c', shader, '-flimit-file']
    expected_error = "glslc: error: argument to '-flimit-file' is missing\n"


@inside_glslc_testsuite('OptionFLimitFile')
class TestFLimitFileMissingFile(expect.ErrorMessageSubstr):
    """Tests -flimit-file without an argument"""

    shader = shader_with_tex_offset(-9);
    glslc_args = ['-c', shader, '-flimit-file', 'i do not exist']
    expected_error_substr = "glslc: error: cannot open input file: 'i do not exist'";


@inside_glslc_testsuite('OptionFLimitFile')
class TestFLimitFileSetsLowerMinTexelOffset(expect.ValidObjectFile):
    """Tests -flimit-file with lower than default argument.  The shader uses that offset."""

    limits_file = File('limits.txt', 'MinProgramTexelOffset -9')
    shader = File('shader.vert', shader_source_with_tex_offset(-9));
    environment = Directory('.', [limits_file, shader])
    glslc_args = ['-c', shader.name, '-flimit-file', limits_file.name]


@inside_glslc_testsuite('OptionFLimitFile')
class TestFLimitFileInvalidContents(expect.ErrorMessage):
    """Tests -flimit-file bad file contents."""

    limits_file = File('limits.txt', 'thisIsBad')
    shader = File('shader.vert', shader_source_with_tex_offset(-9));
    environment = Directory('.', [limits_file, shader])
    glslc_args = ['-c', shader.name, '-flimit-file', limits_file.name]
    expected_error = 'glslc: error: -flimit-file error: invalid resource limit: thisIsBad\n'

## Mesh shading

def mesh_shader_with_params(kwargs):
    """Returns a mesh shader as a FileShader, with given parameters"""
    import sys

    source = """#version 450
#extension {} : enable
      layout(local_size_x={}) in;
      layout(local_size_y={}) in;
      layout(local_size_z={}) in;
      layout(triangles) out;
      layout(max_vertices={}) out;
      layout(max_primitives={}) out;
      layout(triangles) out;
      void main() {{ }}
      """.format(kwargs['extension'],kwargs['x'],kwargs['y'],kwargs['z'],kwargs['max_vert'],kwargs['max_prim'])
    return FileShader(source,".mesh")

def task_shader_with_params(kwargs):
    """Returns a task shader as a FileShader, with given parameters"""
    import sys

    source = """#version 450
#extension {} : enable
      layout(local_size_x={}) in;
      layout(local_size_y={}) in;
      layout(local_size_z={}) in;
      void main() {{ }}
      """.format(kwargs['extension'],kwargs['x'],kwargs['y'],kwargs['z'])
    return FileShader(source,".task")


def meshDefaults(nv_or_ext,show=False):
  result = dict({
    # See Glslang's glslang/ResourceLimits/ResourceLimits.cpp
    'nv': { 'extension': 'GL_NV_mesh_shader', 'x': 32, 'y': 1, 'z': 1, 'max_vert': 256, 'max_prim': 512 },
    'ext': { 'extension': 'GL_EXT_mesh_shader', 'x': 128, 'y': 128, 'z': 128, 'max_vert': 256, 'max_prim': 256 }
  })[nv_or_ext]
  if show:
      import sys
      print(result,file=sys.stderr)
  return result

## GL_NV_mesh_shader

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_NV_X_ok(expect.ValidObjectFile):
    shader = mesh_shader_with_params(meshDefaults('nv'))
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeX_NV 32']

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_NV_X_bad(expect.ErrorMessageSubstr):
    shader = mesh_shader_with_params(meshDefaults('nv'))
    expected_error_substr = "'local_size' : too large, see gl_MaxMeshWorkGroupSizeNV"
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeX_NV 31']

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_NV_Y_ok(expect.ValidObjectFile):
    shader = mesh_shader_with_params(meshDefaults('nv'))
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeY_NV 1']

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_NV_Y_bad(expect.ErrorMessageSubstr):
    d = meshDefaults('nv')
    d['y'] = 3
    shader = mesh_shader_with_params(d)
    expected_error_substr = "'local_size' : too large, see gl_MaxMeshWorkGroupSizeNV"
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeY_NV 2']

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_NV_Z_ok(expect.ValidObjectFile):
    shader = mesh_shader_with_params(meshDefaults('nv'))
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeZ_NV 1']

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_NV_Z_bad(expect.ErrorMessageSubstr):
    d = meshDefaults('nv')
    d['z'] = 3
    shader = mesh_shader_with_params(d)
    expected_error_substr = "'local_size' : too large, see gl_MaxMeshWorkGroupSizeNV"
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeZ_NV 2']

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_NV_MaxVert_ok(expect.ValidObjectFile):
    shader = mesh_shader_with_params(meshDefaults('nv'))
    glslc_args = ['-c', shader, '-flimit= MaxMeshOutputVerticesNV 256']

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_NV_MaxVert_bad(expect.ErrorMessageSubstr):
    shader = mesh_shader_with_params(meshDefaults('nv'))
    expected_error_substr = "'max_vertices' : too large, must be less than gl_MaxMeshOutputVerticesNV"
    glslc_args = ['-c', shader, '-flimit= MaxMeshOutputVerticesNV 255']

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_NV_MaxPrim_ok(expect.ValidObjectFile):
    shader = mesh_shader_with_params(meshDefaults('nv'))
    glslc_args = ['-c', shader, '-flimit= MaxMeshOutputPrimitivesNV 512']

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_NV_MaxPrim_bad(expect.ErrorMessageSubstr):
    shader = mesh_shader_with_params(meshDefaults('nv'))
    expected_error_substr = "'max_primitives' : too large, must be less than gl_MaxMeshOutputPrimitivesNV"
    glslc_args = ['-c', shader, '-flimit= MaxMeshOutputPrimitivesNV 511']

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_NV_X_ok(expect.ValidObjectFile):
    shader = task_shader_with_params(meshDefaults('nv'))
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeX_NV 32']

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_NV_X_bad(expect.ErrorMessageSubstr):
    shader = task_shader_with_params(meshDefaults('nv'))
    expected_error_substr = "'local_size' : too large, see gl_MaxTaskWorkGroupSizeNV"
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeX_NV 31']

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_NV_Y_ok(expect.ValidObjectFile):
    shader = task_shader_with_params(meshDefaults('nv'))
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeY_NV 1']

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_NV_Y_bad(expect.ErrorMessageSubstr):
    d = meshDefaults('nv')
    d['y'] = 3
    shader = task_shader_with_params(d)
    expected_error_substr = "'local_size' : too large, see gl_MaxTaskWorkGroupSizeNV"
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeY_NV 2']

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_NV_Z_ok(expect.ValidObjectFile):
    shader = task_shader_with_params(meshDefaults('nv'))
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeZ_NV 1']

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_NV_Z_bad(expect.ErrorMessageSubstr):
    d = meshDefaults('nv')
    d['z'] = 3
    shader = task_shader_with_params(d)
    expected_error_substr = "'local_size' : too large, see gl_MaxTaskWorkGroupSizeNV"
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeZ_NV 2']

# TODO: Test MaxMeshViewCountNV


## GL_EXT_mesh_shader
## It requires SPIR-V 1.4

s14 = '--target-spv=spv1.4'

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_EXT_X_ok(expect.ValidObjectFile1_4):
    shader = mesh_shader_with_params(meshDefaults('ext'))
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeX_EXT 128', s14]

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_EXT_X_bad(expect.ErrorMessageSubstr):
    shader = mesh_shader_with_params(meshDefaults('ext'))
    expected_error_substr = "'local_size' : too large, see gl_MaxMeshWorkGroupSizeEXT"
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeX_EXT 127', s14]

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_EXT_Y_ok(expect.ValidObjectFile1_4):
    shader = mesh_shader_with_params(meshDefaults('ext'))
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeY_EXT 128', s14]

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_EXT_Y_bad(expect.ErrorMessageSubstr):
    shader = mesh_shader_with_params(meshDefaults('ext'))
    expected_error_substr = "'local_size' : too large, see gl_MaxMeshWorkGroupSizeEXT"
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeY_EXT 127', s14]

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_EXT_Z_ok(expect.ValidObjectFile1_4):
    shader = mesh_shader_with_params(meshDefaults('ext'))
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeZ_EXT 128', s14]

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_EXT_Z_bad(expect.ErrorMessageSubstr):
    shader = mesh_shader_with_params(meshDefaults('ext'))
    expected_error_substr = "'local_size' : too large, see gl_MaxMeshWorkGroupSizeEXT"
    glslc_args = ['-c', shader, '-flimit= MaxMeshWorkGroupSizeZ_EXT 127', s14]

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_EXT_MaxVert_ok(expect.ValidObjectFile1_4):
    shader = mesh_shader_with_params(meshDefaults('ext'))
    glslc_args = ['-c', shader, '-flimit= MaxMeshOutputVerticesEXT 256', s14]

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_EXT_MaxVert_bad(expect.ErrorMessageSubstr):
    shader = mesh_shader_with_params(meshDefaults('ext'))
    expected_error_substr = "'max_vertices' : too large, must be less than gl_MaxMeshOutputVerticesEXT"
    glslc_args = ['-c', shader, '-flimit= MaxMeshOutputVerticesEXT 255', s14]

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_EXT_MaxPrim_ok(expect.ValidObjectFile1_4):
    shader = mesh_shader_with_params(meshDefaults('ext'))
    glslc_args = ['-c', shader, '-flimit= MaxMeshOutputPrimitivesEXT 256', s14]

@inside_glslc_testsuite('OptionFLimit_Mesh')
class TestFLimitMeshShader_EXT_MaxPrim_bad(expect.ErrorMessageSubstr):
    shader = mesh_shader_with_params(meshDefaults('ext'))
    expected_error_substr = "'max_primitives' : too large, must be less than gl_MaxMeshOutputPrimitivesEXT"
    glslc_args = ['-c', shader, '-flimit= MaxMeshOutputPrimitivesEXT 255', s14]

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_EXT_X_ok(expect.ValidObjectFile1_4):
    shader = task_shader_with_params(meshDefaults('ext'))
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeX_EXT 128', s14]

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_EXT_X_bad(expect.ErrorMessageSubstr):
    shader = task_shader_with_params(meshDefaults('ext'))
    expected_error_substr = "'local_size' : too large, see gl_MaxTaskWorkGroupSizeEXT"
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeX_EXT 127', s14]

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_EXT_Y_ok(expect.ValidObjectFile1_4):
    shader = task_shader_with_params(meshDefaults('ext'))
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeY_EXT 128', s14]

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_EXT_Y_bad(expect.ErrorMessageSubstr):
    import sys
    d = meshDefaults('ext',True)
    print("TaskShader_EXT_Y_bad {}".format(str(d)),file=sys.stderr)
    shader = task_shader_with_params(meshDefaults('ext',True))
    expected_error_substr = "'local_size' : too large, see gl_MaxTaskWorkGroupSizeEXT"
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeY_EXT 127', s14]

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_EXT_Z_ok(expect.ValidObjectFile1_4):
    shader = task_shader_with_params(meshDefaults('ext'))
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeZ_EXT 128', s14]

@inside_glslc_testsuite('OptionFLimit_Task')
class TestFLimitTaskShader_EXT_Z_bad(expect.ErrorMessageSubstr):
    shader = task_shader_with_params(meshDefaults('ext'))
    expected_error_substr = "'local_size' : too large, see gl_MaxTaskWorkGroupSizeEXT"
    glslc_args = ['-c', shader, '-flimit= MaxTaskWorkGroupSizeZ_EXT 127', s14]

# TODO: Test MaxMeshViewCountEXT



# Copyright 2018 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader

# A GLSL shader with a separate sampler and texture2D object
GLSL_SHADER_SEPARATE_IMAGE_SAMPLER = """#version 460
  layout (location=0) in vec2 in_UV;
  layout (location=0) out vec4 out_Color;
  layout (set=0,binding=0) uniform sampler u_Sampler;
  layout (set=0,binding=0) uniform texture2D u_Tex;
  void main() {
    out_Color = texture(sampler2D(u_Tex, u_Sampler), in_UV);
  }"""


# An HLSL fragment shader with the usual Texture2D and SamplerState pair
HLSL_SHADER_SEPARATE_IMAGE_SAMPLER = """
  Texture2D u_Tex;
  SamplerState u_Sampler;
  float4 Frag(float2 uv) : COLOR0 {
    return u_Tex.Sample(u_Sampler, uv);
  }"""


@inside_glslc_testsuite('OptionFAutoCombinedImageSampler')
class FAutoCombinedImageSamplerCheckGLSL(expect.ValidAssemblyFileWithSubstr):
    """Tests that the compiler combines GLSL sampler and texture2D objects."""
    shader = FileShader(GLSL_SHADER_SEPARATE_IMAGE_SAMPLER, '.frag')
    glslc_args = ['-S', '-fauto-combined-image-sampler', shader]
    expected_assembly_substr = """%10 = OpTypeImage %float 2D 0 0 0 1 Unknown
         %11 = OpTypeSampledImage %10
%_ptr_UniformConstant_11 = OpTypePointer UniformConstant %11
      %u_Tex = OpVariable %_ptr_UniformConstant_11 UniformConstant"""

@inside_glslc_testsuite('OptionFAutoCombinedImageSampler')
class FAutoCombinedImageSamplerCheckHLSL(expect.ValidAssemblyFileWithSubstr):
    """Tests that the HLSL compiler combines HLSL Texture2D and SamplerState objects into SPIRV SampledImage."""

    shader = FileShader(HLSL_SHADER_SEPARATE_IMAGE_SAMPLER, '.hlsl')
    glslc_args = ['-S', '-fshader-stage=frag', '-fentry-point=Frag', '-fauto-combined-image-sampler', shader]
    expected_assembly_substr = """%14 = OpTypeImage %float 2D 0 0 0 1 Unknown
         %15 = OpTypeSampledImage %14
%_ptr_UniformConstant_15 = OpTypePointer UniformConstant %15
      %u_Tex = OpVariable %_ptr_UniformConstant_15 UniformConstant"""




# Copyright 2015 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader


def opengl_compat_fragment_shader():
    return """#version 330
uniform highp sampler2D tex;
void main() {
  gl_FragColor = texture2D(tex, vec2(0.0, 0.0));
}"""


def opengl_vertex_shader():
    return """#version 330
void main() { int t = gl_VertexID; }"""


def vulkan_vertex_shader():
    return """#version 310 es
void main() { int t = gl_VertexIndex; }"""


def vulkan_compute_subgroup_shader():
    """Returns a compute shader that requires Vulkan 1.1"""
    return """#version 450
              #extension GL_KHR_shader_subgroup_basic : enable
              void main() { subgroupBarrier(); }"""


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqOpenglCompatWithOpenGlCompatShader(expect.ErrorMessageSubstr):
    """Tests that compiling OpenGL Compatibility Fragment shader with
    --target-env=opengl_compat works correctly"""
    shader = FileShader(opengl_compat_fragment_shader(), '.frag')
    glslc_args = ['--target-env=opengl_compat', '-c', shader]
    expected_error_substr = "error: opengl_compat is no longer supported"


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqOpenglWithOpenGlCompatShader(expect.ErrorMessageSubstr):
    """Tests the error message of compiling OpenGL Compatibility Fragment shader
    with --target-env=opengl"""
    shader = FileShader(opengl_compat_fragment_shader(), '.frag')
    glslc_args = ['--target-env=opengl', shader]
    # Glslang does not give a pretty message. Make sure we get an error.
    expected_error_substr = "errors generated"


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqOpenglCompatWithOpenGlVertexShader(expect.ErrorMessageSubstr):
    """Tests that compiling OpenGL vertex shader with --target-env=opengl_compat
    generates valid SPIR-V code"""
    shader = FileShader(opengl_vertex_shader(), '.vert')
    glslc_args = ['--target-env=opengl_compat', '-c', shader]
    expected_error_substr = "error: opengl_compat is no longer supported"


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqOpenglWithOpenGlVertexShader(expect.ValidObjectFile):
    """Tests that compiling OpenGL vertex shader with --target-env=opengl
    generates valid SPIR-V code"""
    shader = FileShader(opengl_vertex_shader(), '.vert')
    glslc_args = ['--target-env=opengl', '-c', shader]


@inside_glslc_testsuite('OptionTargetEnv')
class TestDefaultTargetEnvWithVulkanShader(expect.ValidObjectFile):
    """Tests that compiling a Vulkan-specific shader with a default
    target environment succeeds"""
    shader = FileShader(vulkan_vertex_shader(), '.vert')
    glslc_args = ['-c', shader]


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqVulkanWithVulkan1_0ShaderSucceeds(expect.ValidObjectFile):
    """Tests that compiling a Vulkan-specific Vulkan 1.0 shader succeeds with
    --target-env=vulkan"""
    shader = FileShader(vulkan_vertex_shader(), '.vert')
    glslc_args = ['--target-env=vulkan', '-c', shader]


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqVulkan1_0WithVulkan1_0ShaderSucceeds(expect.ValidObjectFile):
    """Tests that compiling a Vulkan-specific Vulkan 1.0 shader succeeds with
    --target-env=vulkan1.0"""
    shader = FileShader(vulkan_vertex_shader(), '.vert')
    glslc_args = ['--target-env=vulkan1.0', '-c', shader]


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqVulkan1_0WithVulkan1_1ShaderFails(expect.ErrorMessageSubstr):
    shader = FileShader(vulkan_compute_subgroup_shader(), '.comp')
    glslc_args = ['--target-env=vulkan1.0', '-c', shader]
    expected_error_substr = "error: 'subgroup op' : requires SPIR-V 1.3"


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqVulkan1_1WithVulkan1_0ShaderSucceeds(expect.ValidObjectFile1_3):
    shader = FileShader(vulkan_vertex_shader(), '.vert')
    glslc_args = ['--target-env=vulkan1.1', '-c', shader]


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqVulkan1_1WithVulkan1_1ShaderSucceeds(expect.ValidObjectFile1_3):
    shader = FileShader(vulkan_compute_subgroup_shader(), '.comp')
    glslc_args = ['--target-env=vulkan1.1', '-c', shader]

@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqVulkan1_2WithVulkan1_0ShaderSucceeds(expect.ValidObjectFile1_5):
    shader = FileShader(vulkan_vertex_shader(), '.vert')
    glslc_args = ['--target-env=vulkan1.2', '-c', shader]

@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqVulkan1_2WithVulkan1_1ShaderSucceeds(expect.ValidObjectFile1_5):
    shader = FileShader(vulkan_compute_subgroup_shader(), '.comp')
    glslc_args = ['--target-env=vulkan1.2', '-c', shader]

@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqVulkan1_2WithVulkan1_0ShaderSucceeds(expect.ValidObjectFile1_6):
    shader = FileShader(vulkan_vertex_shader(), '.vert')
    glslc_args = ['--target-env=vulkan1.3', '-c', shader]

@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqVulkan1_2WithVulkan1_1ShaderSucceeds(expect.ValidObjectFile1_6):
    shader = FileShader(vulkan_compute_subgroup_shader(), '.comp')
    glslc_args = ['--target-env=vulkan1.3', '-c', shader]

@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqOpenGL4_5WithOpenGLShaderSucceeds(expect.ValidObjectFile):
    shader = FileShader(opengl_vertex_shader(), '.vert')
    glslc_args = ['--target-env=opengl4.5', '-c', shader]


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqOpenGL4_6WithOpenGLShaderFailsUnsupported(expect.ErrorMessageSubstr):
    shader = FileShader(opengl_vertex_shader(), '.vert')
    glslc_args = ['--target-env=opengl4.6', '-c', shader]
    expected_error_substr = "invalid value 'opengl4.6' in '--target-env=opengl4.6'"


# Note: Negative tests are covered in the libshaderc_util unit tests.
# For example, that an OpenGL-specific shader should fail to compile
# for Vulkan, or a Vulkan-specific shader should fail to compile for
# OpenGL.


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqNoArg(expect.ErrorMessage):
    """Tests the error message of assigning empty string to --target-env"""
    shader = FileShader(opengl_vertex_shader(), '.vert')
    glslc_args = ['--target-env=', shader]
    expected_error = ["glslc: error: invalid value ",
                      "'' in '--target-env='\n"]


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvNoEqNoArg(expect.ErrorMessage):
    """Tests the error message of using --target-env without equal sign and
    arguments"""
    shader = FileShader(opengl_vertex_shader(), '.vert')
    glslc_args = ['--target-env', shader]
    expected_error = ["glslc: error: unsupported option: ",
                      "'--target-env'\n"]


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvNoEqWithArg(expect.ErrorMessage):
    """Tests the error message of using --target-env without equal sign but
    arguments"""
    shader = FileShader(opengl_vertex_shader(), '.vert')
    glslc_args = ['--target-env', 'opengl', shader]
    expected_error = ["glslc: error: unsupported option: ",
                      "'--target-env'\n"]


@inside_glslc_testsuite('OptionTargetEnv')
class TestTargetEnvEqWrongArg(expect.ErrorMessage):
    """Tests the error message of using --target-env with wrong argument"""
    shader = FileShader(opengl_vertex_shader(), '.vert')
    glslc_args = ['--target-env=wrong_arg', shader]
    expected_error = ["glslc: error: invalid value ",
                      "'wrong_arg' in '--target-env=wrong_arg'\n"]



# Copyright 2019 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the glslc_test_framework module."""

from glslc_test_framework import get_all_test_methods, get_all_superclasses
import unittest


# Classes to be used in testing get_all_{superclasses|test_methods}()


class Root:

    def check_root(self):
        pass


class A(Root):

    def check_a(self):
        pass


class B(Root):

    def check_b(self):
        pass


class C(Root):

    def check_c(self):
        pass


class D(Root):

    def check_d(self):
        pass


class E(Root):

    def check_e(self):
        pass


class H(B, C, D):

    def check_h(self):
        pass


class I(E):

    def check_i(self):
        pass


class O(H, I):

    def check_o(self):
        pass


class U(A, O):

    def check_u(self):
        pass


class X(U, A):

    def check_x(self):
        pass


class R1:

    def check_r1(self):
        pass


class R2:

    def check_r2(self):
        pass


class Multi(R1, R2):

    def check_multi(self):
        pass


class TestSpirvTestFramework(unittest.TestCase):
    def test_get_all_superclasses(self):
        self.assertEqual(get_all_superclasses(A), [Root])
        self.assertEqual(get_all_superclasses(B), [Root])
        self.assertEqual(get_all_superclasses(C), [Root])
        self.assertEqual(get_all_superclasses(D), [Root])
        self.assertEqual(get_all_superclasses(E), [Root])

        self.assertEqual(get_all_superclasses(H), [Root, B, C, D])
        self.assertEqual(get_all_superclasses(I), [Root, E])

        self.assertEqual(get_all_superclasses(O), [Root, B, C, D, E, H, I])

        self.assertEqual(get_all_superclasses(
            U), [Root, B, C, D, E, H, I, A, O])
        self.assertEqual(get_all_superclasses(
            X), [Root, B, C, D, E, H, I, A, O, U])

        self.assertEqual(get_all_superclasses(Multi), [R1, R2])

    def test_get_all_methods(self):
        self.assertEqual(get_all_test_methods(A), ['check_root', 'check_a'])
        self.assertEqual(get_all_test_methods(B), ['check_root', 'check_b'])
        self.assertEqual(get_all_test_methods(C), ['check_root', 'check_c'])
        self.assertEqual(get_all_test_methods(D), ['check_root', 'check_d'])
        self.assertEqual(get_all_test_methods(E), ['check_root', 'check_e'])

        self.assertEqual(
            get_all_test_methods(H),
            ['check_root', 'check_b', 'check_c', 'check_d', 'check_h'])
        self.assertEqual(get_all_test_methods(
            I), ['check_root', 'check_e', 'check_i'])

        self.assertEqual(
            get_all_test_methods(O), [
                'check_root', 'check_b', 'check_c', 'check_d', 'check_e', 'check_h',
                'check_i', 'check_o'
            ])

        self.assertEqual(
            get_all_test_methods(U), [
                'check_root', 'check_b', 'check_c', 'check_d', 'check_e', 'check_h',
                'check_i', 'check_a', 'check_o', 'check_u'
            ])

        self.assertEqual(
            get_all_test_methods(X), [
                'check_root', 'check_b', 'check_c', 'check_d', 'check_e', 'check_h',
                'check_i', 'check_a', 'check_o', 'check_u', 'check_x'
            ])

        self.assertEqual(
            get_all_test_methods(Multi), ['check_r1', 'check_r2', 'check_multi'])


# Copyright 2018 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader

# A GLSL shader with inputs and outputs explicit locations.
GLSL_SHADER_IO_WITHOUT_LOCATIONS = """#version 310 es
  in vec4 m_in;
  in vec4 m_in1;
  out vec4 m_out;
  out vec4 m_out1;
  void main() {
    m_out = m_in;
    m_out1 = m_in1;
  }"""


# An HLSL fragment shader with inputs and outputs explicit locations.
HLSL_SHADER_IO_WITHOUT_LOCATIONS = """
  float4 Foo(float4 a, float4 b) : COLOR0 {
    return a + b;
  }"""


@inside_glslc_testsuite('OptionFAutoMapLocations')
class MissingLocationsResultsInError(expect.ErrorMessageSubstr):
    """Tests that compilation fails when inputs or outputs have no location."""

    shader = FileShader(GLSL_SHADER_IO_WITHOUT_LOCATIONS, '.vert')
    glslc_args = ['-S', shader]
    expected_error_substr = "SPIR-V requires location for user input/output"


@inside_glslc_testsuite('OptionFAutoMapLocations')
class FAutoMapLocationsGeneratesLocationsCheckInput(expect.ValidAssemblyFileWithSubstr):
    """Tests that the compiler generates locations upon request:  Input 0"""

    shader = FileShader(GLSL_SHADER_IO_WITHOUT_LOCATIONS, '.vert')
    glslc_args = ['-S', shader, '-fauto-map-locations']
    expected_assembly_substr = "OpDecorate %m_in Location 0"


@inside_glslc_testsuite('OptionFAutoMapLocations')
class FAutoMapLocationsGeneratesLocationsCheckOutput0(expect.ValidAssemblyFileWithSubstr):
    """Tests that the compiler generates locations upon request:  Output 0"""

    shader = FileShader(GLSL_SHADER_IO_WITHOUT_LOCATIONS, '.vert')
    glslc_args = ['-S', shader, '-fauto-map-locations']
    expected_assembly_substr = "OpDecorate %m_out Location 0"


# Currently Glslang only generates Location 0.
# See https://github.com/KhronosGroup/glslang/issues/1261
# TODO(dneto): Write tests that check Location 1 is generated for inputs and
# outputs.


# Glslang's HLSL compiler automatically assigns locations inptus and outputs.
@inside_glslc_testsuite('OptionFAutoMapLocations')
class HLSLCompilerGeneratesLocationsCheckInput0(expect.ValidAssemblyFileWithSubstr):
    """Tests that the HLSL compiler generates locations automatically: Input 0."""

    shader = FileShader(HLSL_SHADER_IO_WITHOUT_LOCATIONS, '.hlsl')
    glslc_args = ['-S', '-fshader-stage=frag', '-fentry-point=Foo', shader]
    expected_assembly_substr = "OpDecorate %a Location 0"


@inside_glslc_testsuite('OptionFAutoMapLocations')
class HLSLCompilerGeneratesLocationsCheckOutput(expect.ValidAssemblyFileWithSubstr):
    """Tests that the HLSL compiler generates locations automatically: Output."""

    shader = FileShader(HLSL_SHADER_IO_WITHOUT_LOCATIONS, '.hlsl')
    glslc_args = ['-S', '-fshader-stage=frag', '-fentry-point=Foo', shader]
    expected_assembly_substr = "OpDecorate %_entryPointOutput Location 0"



# Copyright 2015 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
from environment import Directory, File
from glslc_test_framework import inside_glslc_testsuite


@inside_glslc_testsuite('#line')
class TestPoundVersion310InIncludingFile(
    expect.ReturnCodeIsZero, expect.StdoutMatch, expect.StderrMatch):
    """Tests that #line directives follows the behavior of version 310
    (specifying the line number for the next line) when we find a
    #version 310 directive in the including file."""

    environment = Directory('.', [
        File('a.vert', '#version 310 es\n#include "b.glsl"\n'),
        File('b.glsl', 'void main() {}\n')])
    glslc_args = ['-E', 'a.vert']

    expected_stderr = ''
    expected_stdout = \
"""#version 310 es
#extension GL_GOOGLE_include_directive : enable
#line 1 "a.vert"

#line 1 "b.glsl"
void main() { }
#line 3 "a.vert"

"""


@inside_glslc_testsuite('#line')
class TestPoundVersion150InIncludingFile(
    expect.ReturnCodeIsZero, expect.StdoutMatch, expect.StderrMatch):
    """Tests that #line directives follows the behavior of version 150
    (specifying the line number for itself) when we find a #version 150
    directive in the including file."""

    environment = Directory('.', [
        File('a.vert', '#version 150\n#include "b.glsl"\n'),
        File('b.glsl', 'void main() {}\n')])
    glslc_args = ['-E', 'a.vert']

    expected_stderr = ''
    expected_stdout = \
"""#version 150
#extension GL_GOOGLE_include_directive : enable
#line 0 "a.vert"

#line 0 "b.glsl"
void main() { }
#line 2 "a.vert"

"""


@inside_glslc_testsuite('#line')
class TestPoundVersionSyntaxErrorInIncludingFile(expect.ErrorMessageSubstr):
    """Tests that error message for #version directive has the correct
    filename and line number."""

    environment = Directory('.', [
        File('a.vert', '#version abc def\n#include "b.glsl"\n'),
        File('b.glsl', 'void main() {}\n')])
    glslc_args = ['-E', 'a.vert']

    expected_error_substr = [
        "a.vert:1: error: '#version' : must occur first in shader\n",
        "a.vert:1: error: '#version' : must be followed by version number\n",
        "a.vert:1: error: '#version' : bad profile name; use es, core, or "
        "compatibility\n",
    ]


# TODO(antiagainst): now #version in included files results in an error.
# Fix this after #version in included files are supported.
@inside_glslc_testsuite('#line')
class TestPoundVersion310InIncludedFile(expect.ErrorMessageSubstr):
    """Tests that #line directives follows the behavior of version 310
    (specifying the line number for the next line) when we find a
    #version 310 directive in the included file."""

    environment = Directory('.', [
        File('a.vert', '#include "b.glsl"\nvoid main() {}'),
        File('b.glsl', '#version 310 es\n')])
    glslc_args = ['-E', 'a.vert']

    expected_error_substr = [
        "b.glsl:1: error: '#version' : must occur first in shader\n"
    ]


# TODO(antiagainst): now #version in included files results in an error.
# Fix this after #version in included files are supported.
@inside_glslc_testsuite('#line')
class TestPoundVersion150InIncludedFile(expect.ErrorMessageSubstr):
    """Tests that #line directives follows the behavior of version 150
    (specifying the line number for itself) when we find a #version 150
    directive in the included file."""

    environment = Directory('.', [
        File('a.vert', '#include "b.glsl"\nvoid main() {}'),
        File('b.glsl', '#version 150\n')])
    glslc_args = ['-E', 'a.vert']

    expected_error_substr = [
        "b.glsl:1: error: '#version' : must occur first in shader\n"
    ]


@inside_glslc_testsuite('#line')
class TestSpaceAroundPoundVersion310InIncludingFile(
    expect.ReturnCodeIsZero, expect.StdoutMatch, expect.StderrMatch):
    """Tests that spaces around #version & #include directive doesn't matter."""

    environment = Directory('.', [
        File('a.vert', '\t #  \t version   310 \t  es\n#\tinclude  "b.glsl"\n'),
        File('b.glsl', 'void main() {}\n')])
    glslc_args = ['-E', 'a.vert']

    expected_stderr = ''
    expected_stdout = \
"""#version 310 es
#extension GL_GOOGLE_include_directive : enable
#line 1 "a.vert"

#line 1 "b.glsl"
void main() { }
#line 3 "a.vert"

"""


@inside_glslc_testsuite('#line')
class TestSpaceAroundPoundVersion150InIncludingFile(
    expect.ReturnCodeIsZero, expect.StdoutMatch, expect.StderrMatch):
    """Tests that spaces around #version & #include directive doesn't matter."""

    environment = Directory('.', [
        File('a.vert', '  \t #\t\tversion\t   150\t \n# include \t "b.glsl"\n'),
        File('b.glsl', 'void main() {}\n')])
    glslc_args = ['-E', 'a.vert']

    expected_stderr = ''
    expected_stdout = \
"""#version 150
#extension GL_GOOGLE_include_directive : enable
#line 0 "a.vert"

#line 0 "b.glsl"
void main() { }
#line 2 "a.vert"

"""


@inside_glslc_testsuite('#line')
class TestPoundLineWithForcedVersion310(
    expect.ReturnCodeIsZero, expect.StdoutMatch, expect.StderrMatch):
    """Tests that #line directives follows the behavior for the version
    specified via command-line."""

    environment = Directory('.', [
        File('a.vert', '#include "b.glsl"\n'),
        File('b.glsl', 'void main() {}\n')])
    glslc_args = ['-E', '-std=310es', 'a.vert']

    expected_stderr = ''
    expected_stdout = \
"""#extension GL_GOOGLE_include_directive : enable
#line 1 "a.vert"
#line 1 "b.glsl"
void main() { }
#line 2 "a.vert"

"""


@inside_glslc_testsuite('#line')
class TestPoundLineWithForcedVersion150(
    expect.ReturnCodeIsZero, expect.StdoutMatch, expect.StderrMatch):
    """Tests that #line directives follows the behavior for the version
    specified via command-line."""

    environment = Directory('.', [
        File('a.vert', '#include "b.glsl"\n'),
        File('b.glsl', 'void main() {}\n')])
    glslc_args = ['-E', '-std=150', 'a.vert']

    expected_stderr = ''
    expected_stdout = \
"""#extension GL_GOOGLE_include_directive : enable
#line 0 "a.vert"
#line 0 "b.glsl"
void main() { }
#line 1 "a.vert"

"""


@inside_glslc_testsuite('#line')
class TestPoundLineWithForcedDifferentVersion(
    expect.ReturnCodeIsZero, expect.StdoutMatch, expect.StderrMatch):
    """Tests that #line directives follows the behavior for the version
    specified via command-line, even if there is a version specification
    in the source code."""

    environment = Directory('.', [
        File('a.vert', '#version 150\n#include "b.glsl"\n'),
        File('b.glsl', 'void main() {}\n')])
    glslc_args = ['-E', '-std=310es', 'a.vert']

    expected_stderr = ''
    expected_stdout = \
"""#version 150
#extension GL_GOOGLE_include_directive : enable
#line 1 "a.vert"

#line 1 "b.glsl"
void main() { }
#line 3 "a.vert"

"""


@inside_glslc_testsuite('#line')
class TestErrorsFromMultipleFiles(expect.ErrorMessage):
    """Tests that errors from different files have the correct error message
    filename specification."""
    including_file = '''#version 310 es
#include "error.glsl"
int no_return() {}
#include "main.glsl"
'''

    environment = Directory('.', [
        File('a.vert', including_file),
        File('error.glsl', 'int unknown_identifier(int) { return a; }'),
        File('main.glsl', 'void main() {\n  int b = 1.5;\n}')])
    glslc_args = ['-c', 'a.vert']

    expected_error = [
        "error.glsl:1: error: 'a' : undeclared identifier\n",
        "error.glsl:1: error: 'return' : type does not match, or is not "
        "convertible to, the function's return type\n",
        "a.vert:3: error: '' : function does not return a value: no_return\n",
        "main.glsl:2: error: '=' :  cannot convert from ' const float' to "
        "' temp highp int'\n",
        "4 errors generated.\n"]


@inside_glslc_testsuite('#line')
class TestExplicitPoundLineWithPoundInclude(
    expect.ReturnCodeIsZero, expect.StdoutMatch, expect.StderrMatch):
    """Tests that #line works correctly in the presence of #include (which
    itself will generate some #line directives."""
    including_file = '''#version 310 es
#line 10000 "injected.glsl"
int plus1(int a) { return a + 1; }
#include "inc.glsl"
int plus2(int a) { return a + 2; }
#line 55555
#include "main.glsl"
'''

    environment = Directory('.', [
        File('a.vert', including_file),
        File('inc.glsl', 'int inc(int a) { return a + 1; }'),
        File('main.glsl', 'void main() {\n  gl_Position = vec4(1.);\n}')])
    glslc_args = ['-E', 'a.vert']

    expected_stderr = ''
    expected_stdout = '''#version 310 es
#extension GL_GOOGLE_include_directive : enable
#line 1 "a.vert"

#line 10000 "injected.glsl"
int plus1(int a) { return a + 1; }
#line 1 "inc.glsl"
 int inc(int a) { return a + 1; }
#line 10002 "injected.glsl"
 int plus2(int a) { return a + 2; }
#line 55555
#line 1 "main.glsl"
 void main() {
  gl_Position = vec4(1.);
}
#line 55556 "injected.glsl"

'''



# Copyright 2017 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
import os.path
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader


@inside_glslc_testsuite('ErrorNoObject')
class ErrorGeneratesNoObjectFile(expect.NoObjectFile,
                                 expect.NoOutputOnStdout,
                                 expect.ErrorMessageSubstr):
    """Tests that on error, no object file is generated."""

    shader = FileShader('#version 150\nBad', '.frag')
    glslc_args = ['-c', shader]
    expected_error_substr = ['syntax error']


@inside_glslc_testsuite('ErrorNoObject')
class FailureToMakeOutputFileIsErrorWithNoOutputFile(
         expect.NoNamedOutputFiles,
         expect.NoOutputOnStdout,
         expect.ErrorMessageSubstr):
    """Tests that if we fail to make an output file, no file is generated,
    and we have certain error messages."""

    shader = FileShader('#version 150\nvoid main() {}', '.frag')
    bad_file = '/file/should/not/exist/today'
    glslc_args = ['-c', shader, '-o', bad_file]
    expected_output_filenames = [bad_file]
    expected_error_substr = ['cannot open output file']


@inside_glslc_testsuite('ErrorNoObject')
class FailureToMakeOutputFileAsCurrentDirIsErrorWithNoOutputFile(
         expect.NoNamedOutputFiles,
         expect.NoOutputOnStdout,
         expect.ErrorMessageSubstr):
    """Tests that if we fail to make an output file because it is the current
    directory, then no file is generated, and we have certain error messages."""

    shader = FileShader('#version 150\nvoid main() {}', '.frag')
    bad_file = '.'  # Current directory
    glslc_args = ['-c', shader, '-o', bad_file]
    expected_output_filenames = [bad_file]
    expected_error_substr = ['cannot open output file']



# Copyright 2015 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
from glslc_test_framework import inside_glslc_testsuite
from placeholder import FileShader

MINIMAL_SHADER = "#version 140\nvoid main(){}"
# This one is valid GLSL but not valid HLSL.
GLSL_VERTEX_SHADER = "#version 140\nvoid main(){ gl_Position = vec4(1.0);}"
# This one is GLSL but without leading #version.  Should result in
# a parser error when compiled as HLSL.
GLSL_VERTEX_SHADER_WITHOUT_VERSION = "void main(){ gl_Position = vec4(1.0);}"
# This one is valid HLSL but not valid GLSL.
# Use entry point "main" so we don't have to specify -fentry-point
HLSL_VERTEX_SHADER = "float4 main() : SV_POSITION { return float4(1.0); }"

@inside_glslc_testsuite('OptionDashX')
class TestDashXNoArg(expect.ErrorMessage):
    """Tests -x with nothing."""

    glslc_args = ['-x']
    expected_error = [
        "glslc: error: argument to '-x' is missing (expected 1 value)\n",
        'glslc: error: no input files\n']


@inside_glslc_testsuite('OptionDashX')
class TestDashXGlslOnGlslShader(expect.ValidObjectFile):
    """Tests -x glsl on a GLSL shader."""

    shader = FileShader(GLSL_VERTEX_SHADER, '.vert')
    glslc_args = ['-x', 'glsl', '-c', shader]


@inside_glslc_testsuite('OptionDashX')
class TestDashXGlslOnHlslShader(expect.ErrorMessageSubstr):
    """Tests -x glsl on an HLSL shader."""

    shader = FileShader(HLSL_VERTEX_SHADER, '.vert')
    glslc_args = ['-x', 'glsl', '-c', shader]
    expected_error_substr = ["error: #version: Desktop shaders for Vulkan SPIR-V"
                             " require version 140 or higher\n"]


@inside_glslc_testsuite('OptionDashX')
class TestDashXHlslOnGlslShader(expect.ErrorMessageSubstr):
    """Tests -x hlsl on a GLSL shader."""

    shader = FileShader(GLSL_VERTEX_SHADER, '.vert')
    glslc_args = ['-x', 'hlsl', '-c', shader]
    expected_error_substr = ["error: '#version' : invalid preprocessor command\n"]


@inside_glslc_testsuite('OptionDashX')
class TestDashXHlslOnGlslShaderWithoutVertex(expect.ErrorMessageSubstr):
    """Tests -x hlsl on a GLSL shader without leading #version."""

    shader = FileShader(GLSL_VERTEX_SHADER_WITHOUT_VERSION, '.vert')
    glslc_args = ['-x', 'hlsl', '-c', shader]
    expected_error_substr = ["error: 'vec4' : no matching overloaded function found\n"]


@inside_glslc_testsuite('OptionDashX')
class TestDashXWrongParam(expect.ErrorMessage):
    """Tests -x with wrong parameter."""

    shader = FileShader(MINIMAL_SHADER, '.vert')
    glslc_args = ['-x', 'gl', shader]
    expected_error = ["glslc: error: language not recognized: 'gl'\n"]


@inside_glslc_testsuite('OptionDashX')
class TestMultipleDashX(expect.ValidObjectFile):
    """Tests that multiple -x works with a single language."""

    shader = FileShader(GLSL_VERTEX_SHADER, '.vert')
    glslc_args = ['-c', '-x', 'glsl', '-x', 'glsl', shader, '-x', 'glsl']


@inside_glslc_testsuite('OptionDashX')
class TestMultipleDashXMixedLanguages(expect.ValidObjectFile):
    """Tests that multiple -x works with different languages."""

    glsl_shader = FileShader(GLSL_VERTEX_SHADER, '.vert')
    hlsl_shader = FileShader(HLSL_VERTEX_SHADER, '.vert')
    glslc_args = ['-c', '-x', 'hlsl', hlsl_shader,
                  '-x', 'glsl', glsl_shader,
                  '-x', 'hlsl', hlsl_shader,
                  '-x', 'glsl', glsl_shader]


@inside_glslc_testsuite('OptionDashX')
class TestMultipleDashXCorrectWrong(expect.ErrorMessage):
    """Tests -x glsl -x [wrong-language]."""

    shader = FileShader(MINIMAL_SHADER, '.vert')
    glslc_args = ['-x', 'glsl', '-x', 'foo', shader]
    expected_error = ["glslc: error: language not recognized: 'foo'\n"]


@inside_glslc_testsuite('OptionDashX')
class TestMultipleDashXWrongCorrect(expect.ErrorMessage):
    """Tests -x [wrong-language] -x glsl."""

    shader = FileShader(MINIMAL_SHADER, '.vert')
    glslc_args = ['-xbar', '-x', 'glsl', shader]
    expected_error = ["glslc: error: language not recognized: 'bar'\n"]


@inside_glslc_testsuite('OptionDashX')
class TestDashXGlslConcatenated(expect.ValidObjectFile):
    """Tests -xglsl."""

    shader = FileShader(MINIMAL_SHADER, '.vert')
    glslc_args = ['-xglsl', shader, '-c']


@inside_glslc_testsuite('OptionDashX')
class TestDashXWrongParamConcatenated(expect.ErrorMessage):
    """Tests -x concatenated with a wrong language."""

    shader = FileShader(MINIMAL_SHADER, '.vert')
    glslc_args = ['-xsl', shader]
    expected_error = ["glslc: error: language not recognized: 'sl'\n"]


@inside_glslc_testsuite('OptionDashX')
class TestDashXEmpty(expect.ErrorMessage):
    """Tests -x ''."""

    shader = FileShader(MINIMAL_SHADER, '.vert')
    glslc_args = ['-x', '', shader]
    expected_error = ["glslc: error: language not recognized: ''\n"]



# Copyright 2015 The Shaderc Authors. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import expect
from glslc_test_framework import GlslCTest, inside_glslc_testsuite
from placeholder import FileShader


def empty_es_310_shader():
    return '#version 310 es\n void main() {}\n'


@inside_glslc_testsuite('FileExtension')
class VerifyVertExtension(expect.ValidObjectFile):
    """Tests glslc accepts vertex shader extension (.vert)."""

    shader = FileShader(empty_es_310_shader(), '.vert')
    glslc_args = ['-c', shader]


@inside_glslc_testsuite('FileExtension')
class VerifyFragExtension(expect.ValidObjectFile):
    """Tests glslc accepts fragment shader extension (.frag)."""

    shader = FileShader(empty_es_310_shader(), '.frag')
    glslc_args = ['-c', shader]


@inside_glslc_testsuite('FileExtension')
class VerifyTescExtension(expect.ValidObjectFile):
    """Tests glslc accepts tessellation control shader extension (.tesc)."""

    shader = FileShader(
        '#version 440 core\n layout(vertices = 3) out;\n void main() {}',
        '.tesc')
    glslc_args = ['-c', shader]


@inside_glslc_testsuite('FileExtension')
class VerifyTeseExtension(expect.ValidObjectFile):
    """Tests glslc accepts tessellation evaluation shader extension (.tese)."""

    shader = FileShader(
        '#version 440 core\n layout(triangles) in;\n void main() {}', '.tese')
    glslc_args = ['-c', shader]


@inside_glslc_testsuite('FileExtension')
class VerifyGeomExtension(expect.ValidObjectFile):
    """Tests glslc accepts geomtry shader extension (.geom)."""

    shader = FileShader(
        '#version 150 core\n layout (triangles) in;\n'
        'layout (line_strip, max_vertices = 4) out;\n void main() {}',
        '.geom')
    glslc_args = ['-c', shader]


@inside_glslc_testsuite('FileExtension')
class VerifyCompExtension(expect.ValidObjectFile):
    """Tests glslc accepts compute shader extension (.comp)."""

    shader = FileShader(empty_es_310_shader(), '.comp')
    glslc_args = ['-c', shader]


@inside_glslc_testsuite('FileExtension')
class InvalidExtension(expect.ErrorMessage):
    """Tests the error message if a file extension cannot be determined."""

    shader = FileShader('#version 150\n', '.fraga')
    glslc_args = ['-c', shader]
    expected_error = [
        "glslc: error: '", shader,
        "': file not recognized: File format not recognized\n"]
